#!/usr/bin/env python

import argparse
import json
import pathlib
import sys
from distutils.util import strtobool

import pandas as pd
import seaborn as sns
from tqdm import tqdm

import sentspace
import sentspace.utils as utils
from itertools import chain
from functools import reduce, partial
# import numpy as np

import multiprocessing



#     # Plot input data to benchmark data
#     #utils.plot_usr_input_against_benchmark_dist_plots(df_benchmark, sent_embed)

def main(args):
    '''used to run the main pipeline, start to end, depending on the arguments and flags
    '''
    
    utils.io.log(f'SENTSPACE. Received arguments: {args}')

    # Estimate sentence embeddings
    output_dir = sentspace.run_sentence_features_pipeline(args.input_file, stop_words_file=args.stop_words,
                                                          benchmark_file=args.benchmark, process_lexical=args.lexical,
                                                          process_syntax=args.syntax, process_embedding=args.embedding,
                                                          process_semantic=args.semantic,
                                                          output_dir=args.output_dir,
                                                          output_format=args.output_format,
                                                          parallelize=args.parallelize,
                                                          # TODO: return_df or return_path?
                                                          emb_data_dir=args.emb_data_dir)

    with (output_dir/'FINISHED').open('w+') as f:
        pass

if __name__ == "__main__":

    # Parse input
    parser = argparse.ArgumentParser('sentspace', 
                                      usage="""
                                            """
    )

    parser.add_argument('input_file', type=str,
                        help='path to input file or a single sentence. If '
                             'supplying a file, it must be .csv .txt or .xlsx,'
                             ' e.g., example/example.csv')

    # Add an option for a user to include their own stop words
    parser.add_argument('-sw','--stop_words', default=None, type=str, 
                        help='path to delimited file of words to filter out from analysis, e.g., example/stopwords.txt')
    
    # Add an option for a user to choose their benchmark
    parser.add_argument('-b', '--benchmark', type=str,
                        default='sentspace/benchmarks/lexical/UD_corpora_lex_features_sents_all.csv',
                         help='path to csv file of benchmark corpora For example benchmarks/lexical/UD_corpora_lex_features_sents_all.csv')

    # parser.add_argument('--cache_dir', default='.cache', type=str,
    #                  	help='path to directory where results may be cached')

    parser.add_argument('-p', '--parallelize', default=True, type=strtobool, 
                        help='use multiple threads to compute features? '
                             'disable using `-p False` in case issues arise.')

    parser.add_argument('-o', '--output_dir', default='./out', type=str,
                         help='path to output directory where results may be stored')

    parser.add_argument('-of', '--output_format', default='pkl', type=str,
                        choices=['pkl','tsv'])

    # Add an option for a user to choose to not do some analyses. Default is true
    parser.add_argument('-lex','--lexical', type=strtobool, default=False, help='compute lexical features? [False]')
    parser.add_argument('-syn','--syntax', type=strtobool, default=False, help='compute syntactic features? [False]')
    parser.add_argument('-emb','--embedding', type=strtobool, default=False, help='compute high-dimensional sentence representations? [False]')
    parser.add_argument('-sem','--semantic', type=strtobool, default=False, help='compute semantic (multi-word) features? [False]')
    
    parser.add_argument('--emb_data_dir', default='/om/data/public/glove/', type=str,
                         help='path to output directory where results may be stored')
    # parser.add_argument('--cache_dir', default=)

    args = parser.parse_args()	
    main(args)



##############################################################################
##                                                                           ##
## This file is part of ModelBlocks. Copyright 2009, ModelBlocks developers. ##
##                                                                           ##
##    ModelBlocks is free software: you can redistribute it and/or modify    ##
##    it under the terms of the GNU General Public License as published by   ##
##    the Free Software Foundation, either version 3 of the License, or      ##
##    (at your option) any later version.                                    ##
##                                                                           ##
##    ModelBlocks is distributed in the hope that it will be useful,         ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of         ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          ##
##    GNU General Public License for more details.                           ##
##                                                                           ##
##    You should have received a copy of the GNU General Public License      ##
##    along with ModelBlocks.  If not, see <http://www.gnu.org/licenses/>.   ##
##                                                                           ##
###############################################################################

import re
import sys

# a Tree consists of a category label 'c' and a list of child Trees 'ch'
class Tree:

    def __init__(self,c='',ch=[], p=None, l=0, r=0):
        self.c  = c
        self.ch = ch
        self.p = p
        self.l = l
        self.r = r


    # obtain string from tree
    def __str__(self):
        if self.ch == []:
            if not hasattr(self, 'e'):
                return self.c
            else:
                return self.c + '[' + str(self.e) + ']'
        s = '(' + self.c
        if hasattr(self, 'e'):
            s += '[' + str(self.e) + ']'
        for t in self.ch:
            s += ' ' + t.__str__()
        return s + ')'


    def words(self):
        if self.ch == []:
            return [self.c]
        else:
            if type(self.ch[0]) is str:
                return self.ch
            l = []
            for t in self.ch:
                l += t.words()
            return l


    def syncats(self):
        if len(self.ch) == 1 and len(self.ch[0].ch) == 0:
            return [self.c]
        else:
            if type(self.ch[0]) is str:
                return self.ch
            l = []
            for t in self.ch:
                l += t.syncats()
            return l

    # obtain tree from string
    def read(self,s,fIndex=0):
        self.ch = []
        # parse a string delimited by whitespace as a terminal branch (a leaf)
        m = re.search('^ *([^ ()]+) *(.*)',s)
        if m != None:
            (self.c,s) = m.groups()
            self.l = fIndex
            self.r = fIndex
            return s, fIndex+1
        # parse nested parens as a non-terminal branch
        m = re.search('^ *\( *([^ ()]*) *(.*)',s)
        if m != None:
            (self.c,s) = m.groups()
            self.l = fIndex
            # read children until close paren
            while True:
                m = re.search('^ *\) *(.*)',s)
                if m != None:
                    return m.group(1), fIndex
                t = Tree()
                s, fIndex = t.read(s, fIndex)
                self.ch += [t]
                t.p = self
                self.r = t.r
        return ''

    # return the lowest child, but not leaf, that spans the range
    def findBySpan(self, left, right):
        #sys.stderr.write( 'left: ' + str(left) + ' right: ' + str(right) + '\n')
        #sys.stderr.write( 's.left: ' + str(self.l) + ' s.right: ' + str(self.r) + '\n')
        if self.l == 0 and self.r == 0:
            for child in self.ch:
                if type(child) is Tree:
                    child.p = self
            if self.p != None:
                ind = self.p.ch.index(self)
                if ind == 0:
                    self.l = self.p.l
                else:
                    self.l = self.p.ch[ind-1].r
            self.r = self.l + len(self.words())
        #sys.stderr.write( 's.left: ' + str(self.l) + ' s.right: ' + str(self.r) + '\n')
        if left == self.l and right == self.r:
            t = self
            #while len(t.ch) == 1 and len(t.ch[0].ch) > 0:
            #    t = t.ch[0]
            return t
        if left < self.l or right > self.r:
            return None
        for child in self.ch:
            val = child.findBySpan(left, right)
            if val != None:
                return val
        raise 'No node span the range: ' + str(left) + ',' + str(right)


    def findByLeftAndHeight(self, left, height):
        t = self.findBySpan(left, left)
        while height > 0:
            t = t.p
            height -= 1
        return t  
    
    
    def sibling(self):
        if not self.p or len(self.p.ch) == 1:
            return None
        elif self.p.ch[0] == self:
            return self.p.ch[1]
        else:
            return self.p.ch[0]  
    
    def setRefs(self,ctr=0):
        if len(self.ch)==1:
            #self.e = self.ch[0].c[0]+str(ctr)
            m = re.search('\#(.)',self.ch[0].c if len(self.ch[0].ch)==0 else self.ch[0].ch[0].c if len(self.ch[0].ch[0].ch)==0 else self.ch[0].ch[0].ch[0].c)
            if m != None:
                #self.e = 'e' + str(ctr)
                self.e = m.group(1).lower() + str(ctr)
            else:
                #self.e = 'e' + str(ctr)
                self.e = (self.ch[0].c if len(self.ch[0].ch)==0 else self.ch[0].ch[0].c if len(self.ch[0].ch[0].ch)==0 else self.ch[0].ch[0].ch[0].c)[0].lower() + str(ctr)
            ctr += 1
        else:
            #if self.e: self.e = None
            for st in self.ch:
                #if re.match('-lI$',st.c):
                st.e = self.e # by default, inherit parent ref
                ctr = st.setRefs(ctr)
                if re.search('-lC',st.c) != None:
                    self.e = st.e
                else:
                    #if self.e == None and
                    if re.search('-lI',st.c) != None:
                        self.e = st.e
        return ctr

    def getAncestors(self):
        ancestors = []
        p = self  #put itself as the first element in its ancestor list
        while p != None:
            ancestors.append(p)
            p = p.p
        return ancestors
    
    def findArgBoundaries(self, predicateIdx, argHeadwordIdx, allHeadwordIdxs):
        maxProj = self.findMaxProj(predicateIdx, argHeadwordIdx, allHeadwordIdxs)
        return maxProj.l, maxProj.r
    
    def leftBoundary(self):
        l = self
        while len(l.ch) != 1:
            l = l.ch[0]
        return l.e[1:]
    
    def rightBoundary(self):
        r = self
        while len(r.ch) != 1:
            r = r.ch[1]
        return r.e[1:]
    
    def coverOtherArg(self, argHeadwordIdx, allHeadwordIdxs):
        for i in allHeadwordIdxs:
            if i != argHeadwordIdx and i >= self.l and i <= self.r:
                return True
        return False
    
    def findMaxProj(self, predicateIdx, argHeadwordIdx, allHeadwordIdxs):
        predTree = self.treeAt(predicateIdx)
        argTree = self.treeAt(argHeadwordIdx)
        predAncestors = predTree.getAncestors()
        while (not argTree.p in predAncestors) and not argTree.p.coverOtherArg(argHeadwordIdx, allHeadwordIdxs):
            argTree = argTree.p 
        return argTree
    
    def treeAt(self, idx):
        if idx < self.l or idx > self.r:
            raise 'idx is out of range'
        if self.l == self.r and self.l == idx:
            return self
        if idx <= self.ch[0].r:
            return self.ch[0].treeAt(idx)
        if idx >= self.ch[1].l:
            return self.ch[1].treeAt(idx)
        
    def toLatex(self, spaces=1, leafColor=None, refColor=None, guoColor=None, lColor=None):
        color = ("\\" + leafColor) if leafColor else ""
        rColor = ("\\" + refColor) if refColor else ""
        if refColor and self.e:
            ref = " ({" + rColor + " " + self.e + "})"
        else:
            ref = ""
        annotatedCat = self.annotateColors(guoColor, lColor)
        if len(self.ch) == 1:
            return "\lingtree{\TR{" + annotatedCat + ref + "}}{ \TR{ " + color + " " + self.ch[0].c + "} }"
        else:
            s = ""
            for i in range(spaces):
                s += "  "
            spaces += 1
            return "\lingtree{\TR{" + annotatedCat + ref + "}}{\n" + s + self.ch[0].toLatex(spaces, leafColor, refColor, guoColor, lColor) \
                + "\n" + s + self.ch[1].toLatex(spaces, leafColor, refColor, guoColor, lColor) + "\n" + s + "}" 
     
    def annotateColors(self, guoColor, lColor):
        annotatedCat = self.c
        if guoColor:
            annotatedCat = re.sub('(\-[guo][A-Z]*)', '{\\\\' + guoColor + r' \1}', annotatedCat)
        if lColor:
            annotatedCat = re.sub('(\-[l][A-Z])', '{\\\\' + lColor + r' \1}', annotatedCat) 
        return annotatedCat
    



import seaborn as sns
import pandas as pd

def plot_embed(databases):
    features = []
    values = []
    for feat, d in databases.items():
        values.extend(d.values())
        features.extend([feat]*len(d))
    df = pd.DataFrame({'feature': features, 'value': values})
    print('Preparing plots')
    g = sns.FacetGrid(df, col='feature', col_wrap=3,
                      sharex=False, sharey=False)
    g.map(sns.distplot, 'value')
    plt.show()




def plot_bar(labels_plot, func, ylabel, title, yerr_func=None):
    """
    Return bar plot & list of values plotted
    given function to apply for each feature
    """
    x = np.arange(len(labels_plot))
    y = [func(label) for label in labels_plot]
    fig, ax = plt.subplots()
    plt.grid(color='grey', which='both',
             linestyle='-', axis='y', linewidth=0.5)
    if yerr_func:  # error bars
        yerr = [yerr_func(label) for label in labels_plot]
        ax.bar(x, y, yerr=yerr, align='center', alpha=0.5)
    else:
        ax.bar(x, y, align='center', alpha=0.5)
    ax.set(ylabel=ylabel, title=title)
    plt.xticks(x, labels_plot, rotation='vertical')
    return fig, y




def plot_all(result, labels_plot, wordlst_l, save=False, save_path=None):
    """
    Bar plots on NAs in data
    """
    numSentences = len(np.unique(result['Sentence no.']))
    numUniqueWords = len(np.unique(result['Word cleaned']))

    figA, yA = plot_bar(labels_plot, lambda feature: countNA(result[feature]),
                        ylabel=f'No. NA values out of {len(result)} words',
                        title='A. No. NA values for each feature')
    figB, yB = plot_bar(labels_plot, lambda feature: len(uniqueNA(result, feature)),
                        ylabel=f'No. unique NA values out of {len(result)} words ({numUniqueWords} unique words)',
                        title='B. No. unique words with NA values for each feature')
    figC, yC = plot_bar(labels_plot, lambda feature: np.mean(avgNA(result, feature)),
                        ylabel='Mean proportion of NA values within each sentence',
                        # \n Errorbars denote std across sentences'
                        title='C. Mean proportion of words with NA values within each sentence',
                        yerr_func=lambda feature: np.std(avgNA(result, feature)))

    def findProp(v, cutoff=0.5):
        return sum(v >= cutoff)

    figD, yD = plot_bar(labels_plot, lambda feature: findProp(avgNA(result, feature)),
                        ylabel=f'No. sentences out of {numSentences} sentences',
                        title='D. No. sentences with mean proportion of NA values .50 or higher')
    if save:
        for fig, name in zip([figA, figB, figC, figD], ['A', 'B', 'C', 'D']):
            fig.savefig(save_path + name + '.png', bbox_inches='tight')
    return figA, figB, figC, figD


def annotateDecomp(X_decomp, annotation, adjustment_method='skip', x_min=-0.02, x_max=0.02, y_min=-0.01, y_max=0.02, skip_no=2, dims=[0, 1], save=False):
    fig, ax = plt.subplots(figsize=(35, 25))
    # plot all points
    plt.plot(X_decomp[:, dims[0]], X_decomp[:, dims[1]], 'bo')

    plotted_texts = []
    added_indices = [1]

    if adjustment_method == 'simple':
        texts = [plt.text(X_decomp[i, dims[0]], X_decomp[i, dims[1]], annotation[i],
                          ha='center', va='center') for i in range(len(X_decomp[:, 0]))]
        adjust_text(texts)

    if adjustment_method == 'none':
        texts = [plt.text(X_decomp[i, dims[0]] * (1 + 0.01), X_decomp[i, dims[1]] * (1 + 0.02),
                          annotation[i], ha='center', va='center') for i in range(len(X_decomp[:, 0]))]

    if adjustment_method == 'skip':
        for i in range(len(X_decomp[:, 0])):
            if x_min < X_decomp[i, 0] < x_max:
                if y_min < X_decomp[i, 1] < y_max:
                    if i % skip_no == 0:
                        texts = plt.text(
                            X_decomp[i, dims[0]], X_decomp[i, dims[1]], annotation[i], ha='center', va='center')
                        plotted_texts.append(texts)
            else:
                texts = plt.text(
                    X_decomp[i, dims[0]], X_decomp[i, dims[1]], annotation[i], ha='center', va='center')
                plotted_texts.append(texts)
        adjust_text(plotted_texts)

    if adjustment_method == 'distance':  # OBS HEAVY
        x_dists = []
        for k in added_indices:
            x_dist = np.abs(X_decomp[i, 0] - X_decomp[k, 0])
            x_dists.append(x_dist)
            if min(x_dists) > frac:
                y_dists = []
                for g in added_indices:
                    y_dist = np.abs(X_decomp[i, 1] - X_decomp[g, 1])
                    y_dists.append(y_dist)
                    if min(y_dists) > frac:
                        # plot
                        texts = plt.text(
                            X_decomp[i, dims[0]], X_decomp[i, dims[1]], annotation[i], ha='center', va='center')
                        plotted_texts.append(texts)
                        added_indices.append(i)
        adjust_text(plotted_texts)

    ax.spines['top'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['left'].set_visible(False)
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)
    plt.tight_layout()

    if save:
        fig.savefig(save, dpi=240)


def plot_usr_input_against_benchmark_dist_plots(bench_df, usr_df):
    '''
    This function plots the lexical feature data. So it accepts a df of the form that gets saved at the very end of the pipeline.
    bench_df: A df that Greta et al determined would be a benchmark (i.e. a corpus that we passed through the pipeline)
    usr_df: This is the df that the user would like to compare against the benchmark
    '''
    plt.figure(figsize=(17, 10))
    plt.rcParams.update({'font.size': 5})
    plot_number = 1
    binz = 13
    for col in usr_df.columns:
        if 'Sentence no' in col:
            #plot_number = plot_number + 1
            continue
        ax = plt.subplot(4, 5, plot_number)
        bench_df[col].hist(bins=binz, density=True,
                           alpha=0.25, ax=ax, color="skyblue")
        bench_df[col].plot.kde(color="skyblue")
        usr_df[col].hist(bins=binz, density=True,
                         alpha=0.25, ax=ax, color="orange")
        usr_df[col].plot.kde(color="orange", title=col)

        # plot legend
        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles, ['benchmark', 'usr_input'])

        # Go to the next plot for the next loop
        plot_number = plot_number + 1

    plt.show()



import os
import pickle

import numpy as np
import sentspace.utils
from sentspace.utils import io, text
from sentspace.utils.caching import cache_to_mem #, cache_to_disk
from sentspace.utils.misc import merge_lists



# --------- Lexical features

# list of acceptable feature terms to load_databases(...)
# @cache_to_mem
def get_feature_list():
    return ['NRC_Arousal', 'NRC_Valence', 'OSC', 'aoa', 'concreteness', 'lexical_decision_RT',
            'log_contextual_diversity', 'log_lexical_frequency', 'n_orthographic_neighbors', 'num_morpheme',
            'prevalence', 
            'surprisal-3', 'surprisal-1', 'surprisal-2', 'surprisal-4',
            'total_degree_centrality']

def get_feature_list_using_third_party_libraries():
    return ['polysemy', 'num_morpheme_poly']

def get_feature_list_requiring_calculation():
    return ['PMI']


def get_all_features(sentence: 'sentspace.Sentence.Sentence', databases):
    """
    Given list of words, return dict mapping feature to list of feature values
    """
    
    result = {}
    for feature in get_feature_list() + get_feature_list_using_third_party_libraries():
        # we don't want to compute num_morpheme using the dictionary DB by default. 
        # we want to do it only if the polyglot library is unavailable.
        if feature == 'num_morpheme': 
            continue
        computed_feature = get_feature(sentence, feature, databases) 
        # even though we are computing "num_morpheme_poly", we want to
        # record it as "num_morpheme", since the _poly suffix comes from the polyglot library
        # that provides the feature
        if feature == 'num_morpheme_polyglot':
            try:
                import polyglot
            except ImportError:
                feature = 'num_morpheme'
        result[feature] = computed_feature
    return result


def get_feature(sentence: 'sentspace.Sentence.Sentence', feature, databases={}):
    '''
    get specific `feature` for the tokens in `sentence`; fall back to using `lemmas` if necessary
    '''

    def get_feature_(token, lemma, feature):
        """given a `word` and a feature to extract, returns the value of that
            feature for the `word` using available databases

        Args:
            word (str): the token (word) to extract a feature for
            feature (str): name identifier of the feature acc to predefined convention
            databases (dict, in-scope): dictionary of feature --> (word --> feature_value) dictionaries. 
                                        Defaults to {}.

        Returns:
            Any: feature value
        """
        # if the feature is from a database we have on disk
        # database can be a dictionary or an object that implements
        # get(key, default)
        if feature in get_feature_list():
            feature_dict = databases[feature]            
            try:
                return feature_dict[token]
            except KeyError as e:
                try:
                    return feature_dict[lemma]
                except KeyError as e_:
                    return np.nan

        # Other databases we use from libraries we load such as NLTK-Wordnet and Polyglot
        elif feature in get_feature_list_using_third_party_libraries():
            if feature == 'polysemy':
                from nltk.corpus import wordnet
                # first try it with the token itself
                if (synsets := wordnet.synsets(token)):
                    return len(synsets) # TODO does a word's synset include itself?
                # if token is OOV, try again with the lemma
                elif (synsets := wordnet.synsets(lemma)):
                    return len(synsets)
                # otherwise the len of its synset is 1 (itself)
                return 1
            elif feature == 'num_morpheme_poly':
                try:
                    from polyglot.text import Word
                    # try first to obtain # morphemes of the token
                    if (morphed := Word(token, language='en').morphemes):
                        return len(morphed)
                    # otherwise, try using the lemmatized form
                    elif (morphed := Word(lemma, language='en').morphemes):
                        return len(morphed)
                    # if both token and lemma OOV, then return nan? or 1 (i.e. full word is the morpheme?)
                    # but that only means we failed to analyze its morphology, not necessarily that is
                    # *is* a single morpheme
                    return 1 # np.nan
                except ImportError as e:
                    # fall back to simply using a dictionary-based feature
                    # TODO make a note of this somewhere
                    io.log(e.msg, type='WARN')
                    return get_feature_(token, lemma, 'num_morpheme')

        else:
            raise ValueError(f'unable to compute unknown feature `{feature}`')


    features_list = []

    for token, lemma in zip(sentence.tokens, sentence.lemmas):
        features_list += [get_feature_(token, lemma, feature)]

    return features_list


def return_percentile_df(bench_df, usr_df):
    # Initialize df
    perc_df = pd.DataFrame(columns=usr_df.columns)
    # For each sentence get the percentile scores for each feature
    for index, row in usr_df.iterrows():
        temp_df = {}
        # Iterate through the features
        for col in usr_df.columns:
            if col == 'Sentence no.':
                temp_df[col] = row[col]
                continue
            #print(percentileofscore(bench_df[col],row[col]))
            temp_df[col] = percentileofscore(bench_df[col], row[col])
            #pdb.set_trace()
        # Append percentile feature row
        perc_df = perc_df.append(temp_df, ignore_index=True)

    perc_df.drop(columns=['Sentence no.'])
    return perc_df




@cache_to_mem
def load_databases(features='all', path='.feature_database/', 
                   ignore_case=True,
                  ):
    """
    Load dicts mapping word to feature value
    If one feature, provide in list format
    """
    io.log("loading databases with all features")
    databases = {}
    if features == 'all':
        features = get_feature_list()
    for feature in features:
        if not os.path.exists(path+feature+'.pkl'):
            sentspace.utils.s3.load_feature(key=feature+'.pkl')
        with open(path+feature+'.pkl', 'rb') as f:
            d = pickle.load(f)
            # if ignore_case:  # add lowercase version to feature database
            #     for key, val in d.copy().items():
            #         d[str(key).lower()] = val
            databases[feature] = d

    sanity_check_databases(databases)
    io.log("---done--- loading databases with all features")
    return databases


def sanity_check_databases(databases):
    '''
    perform sanity checks upon loading various datasets
    to ensure validity of the loaded data
    '''

    assert databases['NRC_Arousal']['happy'] == 0.735
    assert databases['NRC_Valence']['happy'] == 1
    assert databases['OSC']['happy'] == 0.951549893181384
    assert abs(databases['aoa']['a'] - 2.893384) < 1e-4
    assert databases['concreteness']['roadsweeper'] == 4.85
    # assert abs(databases['imag']['abbey'] - 5.344) < 1e-4
    assert databases['total_degree_centrality']['a'] == 30
    assert databases['lexical_decision_RT']['a'] == 798.917
    assert abs(databases['log_contextual_diversity']['a'] - 3.9234) < 1e-4
    assert abs(databases['log_lexical_frequency']['a'] - 6.0175) < 1e-4
    assert databases['n_orthographic_neighbors']['a'] == 950.59
    assert databases['num_morpheme']['abbreviated'] == 4
    assert abs(databases['prevalence']['a'] - 1.917) < 1e-3
    assert databases['surprisal-3']['beekeeping'] == 10.258



import sentspace.syntax

text = ("When Sebastian Thrun started working on self-driving cars at "
        "Google in 2007, few people outside of the company took him "
        "seriously. “I can tell you very senior CEOs of major American "
        "car companies would shake my hand and turn away because I wasn’t "
        "worth talking to,” said Thrun, in an interview with Recode earlier "
        "this week.")

features = sentspace.syntax.get_features(text, dlt=True, left_corner=True)
print(features.dlt)
print(features.tree)
print(features.left_corner)

# output = utils.pipeline(text, dlt=True)  # === output = utils.pipeline(text, dlt=True, left_corner = False)
# print(output.left_corner)  # is None



import sys
import re

for line in sys.stdin:
  for preterm,term in re.findall( '([^\(\)]+) ([^\(\)]+)', line ):
    term = term.lower()
    for _,formprefix,formsuffix,_,_,lemmaprefix,lemmasuffix in re.findall( '-x([^:|]*:|[^%:|]*)([^%:|]*?)%([^%:|]*)[|]([^%]?)([^-:|]*:|[^-%:|]*)([^-%:|]*?)%([^-%:|]*)', preterm ):
      term = re.sub( '^'+re.escape(formprefix)+'(.*)'+re.escape(formsuffix)+'$', lemmaprefix+'\\1'+lemmasuffix, term.lower() )
    print( term ) #.lower() )



import pandas as pd
import sys
if sys.version_info[0] < 3:
    from StringIO import StringIO
else:
    from io import StringIO


class Feature:
    def __init__(self):
        self.tree = None
        self.dlt = None
        self.left_corner = None


class Tree:
    """Description of a Tree object"""

    def __init__(self, data=None):
        self.raw = data

    def __repr__(self):
        return repr(self.raw)


def rawdatatodf(data, sentence=None, identifier=None):
    df = pd.read_csv(StringIO(data), sep=' ')
    df.rename(columns={'word': 'token'}, inplace=True)
    df['sentence'] = sentence
    df['index'] = identifier
    df = df[['sentence', 'index', *df.columns[:-2]]]
    return df

DLT = LeftCorner = rawdatatodf


# class DLT:
#     """Description of a DLT object"""

#     def __init__(self, data=None):
#         self.pandas_dataframe = pd.read_csv(StringIO(data), sep=' ')
#         self.raw = data

#     def __repr__(self):
#         return repr(self.pandas_dataframe)


# class LeftCorner:
#     """Description of a LeftCorner object"""

#     def __init__(self, data=None):
#         self.pandas_dataframe = pd.read_csv(StringIO(data), sep=' ')
#         self.raw = data

#     def __repr__(self):
#         return repr(self.pandas_dataframe)



import os
from collections import defaultdict
from pathlib import Path
from typing import List, Union

import pandas as pd
import sentspace.lexical
import sentspace.utils
from pandas.core.frame import DataFrame
from sentspace.embedding import utils
from sentspace.utils import io, text
from sentspace.utils.caching import cache_to_disk, cache_to_mem
from tqdm import tqdm


def get_features(sentence: Union[sentspace.Sentence.Sentence, sentspace.Sentence.SentenceBatch],
                 models_and_methods = [({'glove.840B.300d'}, {'mean', 'median'}),
                                       ({'gpt2-xl', 'distilgpt2', 'gpt2'}, {'last'}),
                                       ({'bert-base-uncased', 'aloxatel/mbert', 'roberta-large'}, {'first'}),
                                      ],
                 vocab=None, data_dir=None):
    """get embedding-based features (e.g. avg, min, max, etc.) for sentence.

    Args:
        sentence (str): sentence to get features for
        models_and_methods (`typing.Collection[typing.Tuple[typing.Collection[str], 
                                                            typing.Collection[typing.Union[str, typing.Callable[np.ndarray]]]]]`):
                this collection maps each embeddings/representation source to the aggregation 
                methods that should be applied to it. None matches all the embeddings/representation sources 
                available. Methods can either be strings specifying standard aggregation methods (first, last, mean, median),
                or a function accepting (n, d) array-like objects and returning (d,) 
                array-like objects where d >= 1.                                 
        vocab (set, optional): vocabulary over all sentences that will be processed in this session.
                                 it is recommended for a calling scope to make this available in order
                                 to save processing time of going through all of Glove each time.
                                 In the future, optimizations may be considered, such as, indexing the
                                 byte offset of a particular token in the Glove file for speedy reading.

    Returns:
        [type]: [description]
    """

    token_embeddings = {}

    if any('glove' in model_name for model_names, _ in models_and_methods for model_name in model_names):
        if vocab is None:
            io.log(f'no vocabulary provided in advance. this may take a while. grab some popcorn ^.^', type='WARN')
        
        for model_names, _ in models_and_methods:
            for model_name in model_names:
                if 'glove' in model_name or 'word2vec' in model_name:
                    w2v = utils.load_embeddings(emb_file=f'{model_name}.txt',
                                                vocab=(*sorted(vocab or sentence.tokens),),
                                                data_dir=data_dir)
                    token_embeddings.update(utils.get_word_embeds(sentence, w2v=w2v, model_name=model_name,
                                                                # TODO: infer dims based on supplied w2v !!
                                                                dims=300))
    for model_names, _ in models_and_methods:
        for model_name in model_names:
            if 'glove' in model_name or 'word2vec' in model_name: continue
            token_embeddings.update(utils.get_huggingface_embeds(sentence, model_name=model_name,
                                                                layers=None)
                                )

    content_word_filter = lambda i, token: sentence.content_words[i]
    filters = {'content_words': content_word_filter}
    
    pooled_embeddings = defaultdict(dict)
    
    for model_names, methods in models_and_methods:
        d = utils.pool_sentence_embeds(sentence, token_embeddings, keys=model_names,
                                       filters=filters, methods=methods)
        # d: model_name -> method_name -> pooled_repr
        for model_name in d:
            pooled_embeddings[model_name].update(d[model_name])

    return {
        'index': sentence.uid,
        'sentence': str(sentence),
        'filters': ','.join(filters.keys()),
        
        'features': pooled_embeddings,
    }


    # # Read in benchmark data
    # df_benchmark = pd.read_csv(benchmark_file)

    # # Return percentile per sentence for each
    # percentile_df = utils.return_percentile_df(df_benchmark, sent_embed)
    # print('Writing percentiles')
    # percentile_df.to_csv(bench_perc_out_path, index=False)



import importlib
import boto3
import os
import sys
from botocore import UNSIGNED
from botocore.config import Config
from tqdm import tqdm

class _S3Storage():
    """
    load pickles
    """

    _NO_SIGNATURE = Config(signature_version=UNSIGNED)

    def __init__(self, *args, key, bucket='sentspace-databases', region='us-east-1', **kwargs):
        super(_S3Storage, self).__init__(*args, **kwargs)
        self._key = key
        self._bucket = bucket
        self._region = region
        self._local_root_dir = os.path.join(os.getcwd(),'.feature_database/')
        os.makedirs(self._local_root_dir, exist_ok=True)
        self._retrieve()

    #is the relevant file here?
    def _retrieve(self):
        key=self._key
        dir=self._local_root_dir
        local_path = os.path.join(dir, key)
        if not os.path.isfile(local_path):
            self._download_file(key, local_path)

    #download it from s3 bucket
    def _download_file(self, key, local_path):
        print(f"Downloading {key} to {local_path}")
        s3 = boto3.resource('s3', region_name=self._region, config=self._NO_SIGNATURE)
        obj = s3.Object(self._bucket, key)
        with tqdm(total=obj.content_length, unit='B', unit_scale=True, desc=key, file=sys.stdout) as progress_bar:
            def progress_hook(bytes_amount):
                progress_bar.update(bytes_amount)

            obj.download_file(local_path, Callback=progress_hook)

    def save(self, result, function_identifier):
        raise NotImplementedError("can only load from S3, but not save")

load_feature = _S3Storage



'''
    ### Sentspace 0.0.1 (C) 2020-2021 EvLab <evlab.mit.edu>, MIT BCS. All rights reserved.

    Homepage: https://github.com/aalok-sathe/sentspace

    Authors (reverse alphabet.; insert `@` symbol for valid email):
    
    - Greta Tuckute `<gretatu % mit.edu>`
    - Aalok Sathe `<asathe % mit.edu>`
    - Alvince Pongos `<apongos % mit.edu>`
    - Josef Affourtit `<jaffourt % mit.edu>`

    Please contact any of the following with questions about the code or license.
    
    - Aalok Sathe `<asathe % mit.edu>` 
    - Greta Tuckute `<gretatu % mit.edu>`

    .. include:: ../README.md
'''

# __pdoc__ = {'semantic': False,
#             }


from collections import defaultdict
from pathlib import Path

import sentspace.utils as utils
import sentspace.syntax as syntax
import sentspace.lexical as lexical
import sentspace.embedding as embedding

from sentspace.Sentence import Sentence

import pandas as pd
from functools import reduce 
from itertools import chain
from tqdm import tqdm


def run_sentence_features_pipeline(input_file: str, stop_words_file: str = None,
                                   benchmark_file: str = None, output_dir: str = None,
                                   output_format: str = None, batch_size: int = 2_000,
                                   process_lexical: bool = False, process_syntax: bool = False,
                                   process_embedding: bool = False, process_semantic: bool = False,
                                   parallelize: bool = True,
                                   #
                                   emb_data_dir: str = None) -> Path:
    """
    Runs the full sentence features pipeline on the given input according to
    requested submodules (currently supported: `lexical`, `syntax`, `embedding`,
    indicated by boolean flags).
        
    Returns an instance of `Path` pointing to the output directory resulting from this
    run of the full pipeline. The output directory contains Pickled or TSVed pandas 
    DataFrames containing the requested features.


    Args:
        input_file (str): path to input text file containing sentences
                            one per line [required]
        stop_words_file (str): path to text file containing stopwords to filter
                                out, one per line [optional]
        benchmark_file (str): path to a file containing a benchmark corpus to
                                compare the current input against; e.g. UD [optional]
        
        {lexical,syntax,embedding,semantic,...} (bool): compute submodule features? [False]
    """

    # lock = multiprocessing.Manager().Lock()

    # create output folder
    utils.io.log('creating output folder')
    output_dir = utils.io.create_output_paths(input_file,
                                              output_dir=output_dir,
                                              stop_words_file=stop_words_file)
    config_out = (output_dir / 'this_session_log.txt')
    # with config_out.open('a+') as f:
    #     print(args, file=f)

    utils.io.log('reading input sentences')
    sentences = utils.io.read_sentences(input_file, stop_words_file=stop_words_file)
    utils.io.log('---done--- reading input sentences')


    for part, sentence_batch in enumerate(tqdm(utils.io.get_batches(sentences, batch_size=batch_size), 
                                               desc='processing batches')):
        sentence_features_filestem = f'sentence-features_part{part:0>4}'
        token_features_filestem = f'token-features_part{part:0>4}'


        ################################################################################
        #### LEXICAL FEATURES ##########################################################
        ################################################################################
        if process_lexical:
            utils.io.log('*** running lexical submodule pipeline')
            _ = lexical.utils.load_databases(features='all')

            if parallelize:
                lexical_features = utils.parallelize(lexical.get_features, sentence_batch,
                                                    wrap_tqdm=True, desc='Lexical pipeline')
            else:
                lexical_features = [lexical.get_features(sentence)
                                    for _, sentence in enumerate(tqdm(sentence_batch, desc='Lexical pipeline'))]

            lexical_out = output_dir / 'lexical'
            lexical_out.mkdir(parents=True, exist_ok=True)
            utils.io.log(f'outputting lexical token dataframe to {lexical_out}')

            # lexical is a special case since it returns dicts per token (rather than per sentence)
            # so we want to flatten it so that pandas creates a sensible dataframe from it.
            token_df = pd.DataFrame(chain.from_iterable(lexical_features))

            if output_format == 'tsv':
                token_df.to_csv(lexical_out / f'{token_features_filestem}.tsv', sep='\t', index=True)
                token_df.groupby('sentence').mean().to_csv(lexical_out / f'{sentence_features_filestem}.tsv', sep='\t', index=True)
            elif output_format == 'pkl':
                token_df.to_pickle(lexical_out / f'{token_features_filestem}.pkl.gz', protocol=5)
                token_df.groupby('sentence').mean().to_pickle(lexical_out / f'{sentence_features_filestem}.pkl.gz', protocol=5)
            else:
                raise ValueError(f'output format {output_format} not known')

            utils.io.log(f'--- finished lexical pipeline')


        ################################################################################
        #### SYNTAX FEATURES ###########################################################
        ################################################################################
        if process_syntax:
            utils.io.log('*** running syntax submodule pipeline')

            # as an exception, we do *not* parallelize syntax since the backend server is somehow unable to handle
            # multiple requests :(
            syntax_features = [syntax.get_features(sentence._raw, dlt=True, left_corner=True, identifier=sentence.uid)
                                                                        # !!! TODO:DEBUG
                            for i, sentence in enumerate(tqdm(sentences, desc='Syntax pipeline'))]

            syntax_out = output_dir / 'syntax'
            syntax_out.mkdir(parents=True, exist_ok=True)

            # put all features in the sentence df except the token-level ones
            token_syntax_features = {'dlt', 'leftcorner'}
            sentence_df = pd.DataFrame([{k: v for k, v in feature_dict.items() if k not in token_syntax_features}
                                        for feature_dict in syntax_features], index=[s.uid() for s in sentences])

            # output gives us dataframes corresponding to each token-level feature. we need to combine these
            # into a single dataframe
            # we use functools.reduce to apply the pd.concat function to all the dataframes and join dataframes
            # that contain different features for the same tokens
            token_dfs = [reduce(lambda x, y: pd.concat([x, y], axis=1, sort=False),
                                (v for k, v in feature_dict.items() if k in token_syntax_features))
                        for feature_dict in syntax_features]

            # by this point we have merged dataframes with tokens along a column (rather than just a sentence)
            # now we need to stack them on top of each other to have all tokens across all sentences in a single dataframe
            token_df = reduce(lambda x, y: pd.concat([x, y], ignore_index=True), token_dfs)
            token_df = token_df.loc[:, ~token_df.columns.duplicated()]

            utils.io.log(f'outputting syntax dataframes to {syntax_out}')
            if output_format == 'tsv':
                sentence_df.to_csv(syntax_out / f'{sentence_features_filestem}.tsv', sep='\t', index=True)
                token_df.to_csv(syntax_out / f'{token_features_filestem}.tsv', sep='\t', index=False)
            elif output_format == 'pkl':
                sentence_df.to_pickle(syntax_out / f'{sentence_features_filestem}.pkl.gz', protocol=5)
                token_df.to_pickle(syntax_out / f'{token_features_filestem}.pkl.gz', protocol=5)

            utils.io.log(f'--- finished syntax pipeline')

        # Calculate PMI
        # utils.GrabNGrams(sent_rows,pmi_paths)
        # utils.pPMI(sent_rows, pmi_paths)
        # pdb.set_trace()


        ################################################################################
        #### EMBEDDING FEATURES ########################################################
        ################################################################################            
        if process_embedding:
            utils.io.log('*** running embedding submodule pipeline')

            models_and_methods = [
                        # ({'glove.840B.300d'}, {'mean', 'median'}), 
                        # 'distilgpt2',
                        ({'gpt2-xl'}, {'last'}),
                        ({'bert-base-uncased'}, {'first'}),
                     ]
            
            vocab = None
            # does any of the
            if any('glove' in model or 'word2vec' in model for models, _ in models_and_methods for model in models):
                # get a vocabulary across all sentences given as input
                # as the first step, remove any punctuation from the tokens
                stripped_tokens = utils.text.strip_words(chain(*[s.tokens for s in sentences]), method='punctuation')
                # assemble a set of unique tokens
                vocab = set(stripped_tokens)
                # make a spurious function call so that loading glove is cached for subsequent calls
                # TODO allow specifying which glove/w2v version 
                _ = embedding.utils.load_embeddings(emb_file='glove.840B.300d.txt',
                                                    vocab=(*sorted(vocab),),
                                                    data_dir=emb_data_dir)

            if False and parallelize:
                embedding_features = utils.parallelize(embedding.get_features, 
                                                       sentences, models=models, 
                                                       vocab=vocab, data_dir=emb_data_dir,
                                                       wrap_tqdm=True, desc='Embedding pipeline')
            else:
                embedding_features = [embedding.get_features(sentence, models_and_methods=models_and_methods, 
                                                             vocab=vocab, data_dir=emb_data_dir)
                                      for i, sentence in enumerate(tqdm(sentences, desc='Embedding pipeline'))]

            # a misc. stat being computed that needs to be handled better 
            # "no" means no. the stat below is counting how many sentences have NO content words (not to be confused with num. content words)
            no_content_words = len(sentences)-sum(any(s.content_words) for s in sentences)

            utils.io.log(f'sentences without any content words: {no_content_words}/{len(sentences)}; {no_content_words/len(sentences):.2f}')

            embedding_out = output_dir / 'embedding'
            embedding_out.mkdir(parents=True, exist_ok=True)
            
            
            # now we want to output stuff from embedding_features (which is returned by the embedding pipeline)
            # into nicely formatted dataframes.
            # the structure of what is returned by the embedding pipeline is like so:
            #   gpt2-xl:
            #       last: [...] flat multiindexed Pandas series with (layer, dim) as the two indices
            #       mean: 
            #   glove:
            #       mean: [...] flat multiindexed Pandas series with trivially a single layer and 300d, so (1, 300) as the two indices
            # etc.

            # a set of all the models in use
            all_models_methods = {model_name: feature_dict['features'][model_name].keys() 
                                  for feature_dict in embedding_features 
                                    for model_name in feature_dict['features']}

            print(all_models_methods)

            # we want to output BY MODEL
            for model_name in all_models_methods:
                # and BY METHOD
                for method in all_models_methods[model_name]:
                    # each `feature_dict` corresponds to ONE sentence

                    collected = []
                    for feature_dict in embedding_features:

                        # all the keys that contain information such as the sentence, UID, filters used etc,
                        # except for the actual representations obtained from various models.
                        # we need to know this so we can package all this information together with the outputs by model and method
                        metadata_keys = {*feature_dict.keys()} - {'features'} # setminus operator
                        # make a copy of the feature_dict for this sentence excluding the representations themselves
                        meta_df = {key: feature_dict[key] for key in metadata_keys}
                        meta_df.update({'model_name': model_name, 'aggregation': method})
                        meta_df = pd.DataFrame(meta_df, index=[feature_dict['index']])
                        meta_df.columns = pd.MultiIndex.from_product([['metadata'], meta_df.columns, ['']])

                        # model_name -> method -> reprs
                        pooled_reprs = feature_dict['features']
                        flattened_repr = pooled_reprs[model_name][method]

                        collected += [pd.concat([meta_df, flattened_repr], axis=1)]

                    # create further subdirectories by model and aggregation method
                    (embedding_out / model_name / method).mkdir(parents=True, exist_ok=True)

                    sentence_df = pd.concat(collected, axis=0)

                    utils.io.log(f'outputting embedding dataframes for {model_name}-{method} to {embedding_out}')
                    if output_format == 'tsv':
                        sentence_df.to_csv(embedding_out / model_name / method / f'{sentence_features_filestem}.tsv', sep='\t', index=True)
                        # token_df.to_csv(embedding_out / f'{token_features_filestem}.tsv', sep='\t', index=False)
                    elif output_format == 'pkl':
                        sentence_df.to_pickle(embedding_out / model_name / method / f'{sentence_features_filestem}.pkl.gz', protocol=5)
                        # token_df.to_pickle(embedding_out / f'{token_features_filestem}.pkl.gz', protocol=5)


            utils.io.log(f'--- finished embedding pipeline')

        # Plot input data to benchmark data
        #utils.plot_usr_input_against_benchmark_dist_plots(df_benchmark, sent_embed)

        if process_semantic:
            pass


    ################################################################################
    #### \end{run_sentence_features_pipeline} ######################################
    ################################################################################
    return output_dir



import json
import os
import sentspace
import subprocess
import time
import urllib
from pathlib import Path
from urllib import request
import http

from nltk.corpus import wordnet

from sentspace.utils import caching, io, text


def path_decorator(func):
    """Decorator that changes to and from the directory containing scripts
        for running DLT and Left Corner metrics before and after a function
        call respectively.
    """

    def wrapped(*args, **kwargs):
        ''' function that changes the directory to an expected directory,
            executes original function with supplied args, and changes
            back to the same directory we started from
        '''
        previous_pwd = os.getcwd()
        target = Path(__file__)
        os.chdir(str(target.parent))
        result = func(*args, **kwargs)
        os.chdir(previous_pwd)
        return result.decode('utf-8').strip()

    return wrapped



def get_content_ratio(is_content_pos_tag: tuple):
    """
    given boolean list corresponding to a token being a content word, 
    calculate the content ratio
    """
    return sum(is_content_pos_tag) / len(is_content_pos_tag)


def get_pronoun_ratio(pos_tags: tuple):
    """
    Given sentence calculate the pronoun ratio
    """
    pronoun_tags = {'PRP', 'PRP$', 'WP', 'WP$'}
    return sum(tag in pronoun_tags for tag in pos_tags) / len(pos_tags)


# @cache_to_mem
def get_is_content(taglst: tuple, content_pos=(wordnet.ADJ, wordnet.VERB, wordnet.NOUN, wordnet.ADV)):
    """
    Given list of POS tags, return list of 1 - content word, 0 - not content word
    """
    return tuple(int(text.get_wordnet_pos(tag) in content_pos) for tag in taglst)


@path_decorator
def tokenize(raw):
    cmd = ['bash', 'tokenize.sh', raw.strip()]
    # io.log(f'calling tokenizer like so: `{cmd}`')
    tokens = subprocess.check_output(cmd)
    # io.log(f'---done--- tokenizer returned output like so: `{tokens}`')
    return tokens



@path_decorator
# @caching.cache_to_disk
def compute_trees(sentence, server_url='http://localhost:8000/fullberk'):

    # data = f'{{ "sentence": "{sentence}" }}'
    data = {"sentence": sentence}
    r = request.Request(server_url, data=bytes(json.dumps(data), 'utf-8'), method='GET',
                        headers={'Content-Type': 'application/json'})
    retries = 3
    for attempt in range(1, retries+1):
        try:
            with request.urlopen(r) as rq:
                response = rq.read()

        except urllib.error.HTTPError as e:
            io.log(f'encountered HTTPError on sentence [{sentence}]; attempt {attempt}/{retries}; are you sure the server is running? https://github.com/aalok-sathe/berkeley-interact for instructions',
                type='ERR')
            print(e)
            time.sleep(6)
            continue
            exit(1)

        except urllib.error.URLError as e:
            io.log(f'encountered URLError on sentence [{sentence}]; attempt {attempt}/{retries}', type='ERR')
            print(e)
            time.sleep(6)
            continue
            exit(1)

        except http.client.RemoteDisconnected as e:
            io.log(f'encountered http.client.RemoteDisconnected on sentence [{sentence}]; attempt {attempt}/{retries}', type='ERR')
            print(e)
            time.sleep(6)
            continue
            exit(1)
        
        break
    else:
        raise RuntimeError(f'failed to process [{sentence}] after {retries} retries')


    cmd = ['bash', 'postprocess_trees.sh', response]
    # fallback to manually initializing parser
    # cmd = ['bash', 'parse_trees.sh', tokens]
    trees = subprocess.check_output(cmd)
    return trees


@path_decorator
def compute_feature(feature, trees):
    cmd = ['bash', feature, trees]
    try:
        completed = subprocess.run(cmd, check=True, capture_output=True)
    except  subprocess.CalledProcessError as e:
        print('ERROR', e.output, e.returncode, sep='\n')
        exit()
    return completed.stdout
    # out = subprocess.check_output(cmd, stderr=subprocess.DEVNULL)
    # return out




from genericpath import exists
import os
import pickle
import textwrap
import datetime # from datetime import date
from hashlib import sha1
from pathlib import Path
from sys import stderr, stdout
import pandas as pd
import numpy as np
import sentspace.utils
from sentspace.utils.caching import cache_to_disk, cache_to_mem
from sentspace.utils.s3 import load_feature
from tqdm import tqdm
from sentspace.Sentence import Sentence


# from sentspace.utils.sanity_checks import sanity_check_databases

def dump_features(): pass

def create_output_paths(input_file:str, output_dir:str, calling_module=None, stop_words_file:str=None) -> Path:
    """Creates an output directory structure to output the results of the pipeline based on the supplied input file

    Args:
        input_file (str): [description]
        output_dir (str): [description]
        calling_module ([type], optional): [description]. Defaults to None.
        stop_words_file (str, optional): [description]. Defaults to None.

    Returns:
        Path: [description]
    """    
    output_dir = Path(output_dir)
    # output will be organized based on its respective input file,
    # together with a hash of the contents (for to avoid the case where two files are 
    # named similarly to one another but contain different contents)
    out_file_name = '.'.join(os.path.basename(input_file).split('.')[:-1])

    output_dir /= out_file_name
    output_dir /= sentspace.utils.md5(input_file)
    output_dir.mkdir(parents=True, exist_ok=True)

    with (output_dir / 'md5.txt').open('w') as f:
        f.write(f'{Path(input_file).resolve()} md5sum:\t' + sentspace.utils.md5(input_file))
    with (output_dir / 'run_history.txt').open('a+') as f:
        f.write(datetime.datetime.now().strftime('run: %Y-%m-%d %H:%M:%S %Z\n'))

    output_dir /= calling_module or '' # to subdir it by "lexical" or "syntax" etc.
    # output_dir /= date.today().strftime('run_%Y-%m-%d')
    output_dir.mkdir(parents=True, exist_ok=True)

    return output_dir


def read_sentences(filename: str, stop_words_file: str = None):
    """reads sentences from a file, one per line, filtering
        for stopwords

    Args:
        filename (str): path to input file containing sentences
        stop_words_file (str, optional): path to file containing stopwords

    Returns:
        list[list]: list of the tokens from each sentences, nested as a list
        list: list of sentences
    """

    if stop_words_file:
        raise NotImplementedError()
        stop_words = set(np.loadtxt(stop_words_file, delimiter='\t', unpack=False, dtype=str))

    if filename.endswith('.txt'):
        UIDs = []
        sentences = []
        with open(filename, 'r') as f:
            UID_prefix = f'{filename[-8:]:#>10}' + '_' + sentspace.utils.md5(filename)[-5:]
            for i, line in enumerate(f):

                uid = UID_prefix + '_' + f'{len(UIDs):0>5}'
                UIDs += [uid]
                
                s = Sentence(line, uid)
                if s: 
                    sentences.append(s)

        return sentences

    elif filename.endswith('.pkl'):
        df = pd.read_pickle(filename)
    elif filename.endswith('.csv'):
        df = pd.read_csv(filename, sep=',')
    elif filename.endswith('.tsv'):
        df = pd.read_csv(filename, sep='\t')
    else:
        raise ValueError('unknown type of file supplied (must be txt/pkl/csv/tsv. if pickle, must be a dataframe object)')

    # try to figure out what to use as a unique identifier for the sentence
    if 'corpora_identifier' in df.columns:
        UIDs = df['corpora_identifier'].tolist()
    elif 'index' in df.columns:
        UIDs = df['index'].tolist()
    else:
        try: 
            UIDs = df.index.tolist()
        except AttributeError:
            raise('does your dataframe have a unique index for each sentence?')

    sentences = [Sentence(raw, uid) for raw, uid in zip(df['sentence'].tolist(), UIDs)]
    return sentences


def get_batches(iterable, batch_size:int):
    """
    splits iterable into batches of size batch_size
    """
    batch = []
    for i, item in enumerate(iterable):
        batch.append(item)
        if (i + 1) % batch_size == 0:
            yield batch
            batch = []
    if batch:
        yield batch


def log(message, type='INFO'):
    
    class T:
        HEADER = '\033[95m'
        OKBLUE = '\033[94m'
        OKCYAN = '\033[96m'
        OKGREEN = '\033[92m'
        WARNING = '\033[93m'
        FAIL = '\033[91m'
        ENDC = '\033[0m'
        BOLD = '\033[1m'
        UNDERLINE = '\033[4m'

    if type == 'INFO':
        c = T.OKCYAN
    elif type == 'EMPH':
        c = T.OKGREEN
    elif type == 'WARN':
        c = T.BOLD + T.WARNING
    elif type == 'ERR':
        c = '\n' + T.BOLD + T.FAIL
    else:
        c = T.OKBLUE

    timestamp = f'{sentspace.utils.time() - sentspace.utils.START_TIME():.2f}s'
    lines = textwrap.wrap(message+T.ENDC,
                          width=120, 
                          initial_indent = c + '%'*4 + f' [{type} @ {timestamp}] ', 
                          subsequent_indent='.'*20+' ')
    tqdm.write('\n'.join(lines), file=stderr)
    # print(*lines, sep='\n', file=stderr)





import concurrent.futures
import hashlib
import math
import pdb
import pickle
from functools import partial
from itertools import chain
from time import time

# import seaborn as sns
import nltk
# import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import scipy.io as sio
import scipy.spatial.distance as ssd
from nltk import pos_tag
from scipy.stats import percentileofscore, zscore
from tqdm import tqdm

# from zs import ZS

_START_TIME = time()
def START_TIME(): return _START_TIME

# lemmas=WordNetLemmatizer()


# def import_franklin_sentences_set3(filename):
#     f2g = []
#     with open(filename, 'r') as file:
#         for line in file:
#             if line.startswith('passage') or line == '\n':
#                 pass
#             else:
#                 f2g.append(line.strip('\n').split())
                
#     return f2g

# def import_data(filename, dtype=None):
#     """
#     Import text file with \n after each line.
#     Pre-computed GloVe vectors can be loaded as:
#         glove_embed = import_data('../glove/vectors_243sentences.txt', dtype=lambda x: float(x))

#     """
#     f1g = []
#     with gzip.open(filename+'.zip', 'r') as file:
#         for line in file:
#             tokens = line.split()
#             if dtype:
#                 tokens = [dtype(token) for token in tokens]
#             f1g.append(tokens)
#     return f1g

# def get_wordlst(f1g):
#     """
#     Given list of sentences (each a list of tokens), return single list of tokens
#     """
#     wordlst = []
#     for sentence in f1g:
#         for word in sentence:
#             wordlst.append(word)
#     return wordlst

# def get_sent_num_passsage(f1g, lplst):
#     """
#     Given list of passage no. for each sentence, return sentence no. within passage (for each word)
#     """
#     sent_num = 0
#     snplst = []
#     current_label = lplst[0]
#     for sentence, label in zip(f1g, lplst):
#         sent_num += 1
#         if label != current_label:
#             sent_num = 1
#             current_label = label
#         for word in sentence:
#             snplst.append(sent_num)
#     return snplst



# download NLTK data if not already downloaded
def download_nltk_resources():
    for category, nltk_resource in [('taggers', 'averaged_perceptron_tagger'), 
                                    ('corpora', 'wordnet'),
                                    # ('tokenizers', 'punkt')
                                    ]:
        try:
            nltk.data.find(category+'/'+nltk_resource)
        except LookupError as e:
            try:
                nltk.download(nltk_resource)
            except FileExistsError:
                pass

def md5(fname) -> str:
    '''generates md5sum of the contents of fname
        fname (str): path to file whose md5sum we want
    '''
    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()

def sha1(ob):
    ob_repr = repr(ob)
    hash_object = hashlib.sha1()
    hash_object.update(ob_repr.encode('utf-8'))
    return hash_object.hexdigest()


def parallelize(function, *iterables, wrap_tqdm=True, desc='', **kwargs):
    """parallelizes a function by calling it on the supplied iterables and (static) kwargs.
       optionally wraps in tqdm for progress visualization 

    Args:
        function ([type]): [description]
        wrap_tqdm (bool, optional): [description]. Defaults to True.
        desc ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    """    
    partialfn = partial(function, **kwargs)
    with concurrent.futures.ProcessPoolExecutor() as executor:
        if wrap_tqdm:
            return [*tqdm(executor.map(partialfn, *iterables), total=len(iterables[0]), desc='[parallelized] '+desc)]
        return executor.map(partialfn, *iterables)

# this might be data-dependent
def load_passage_labels(filename):
    """
    Given .mat file, load and return list of passage no. for each sentence
    """
    labelsPassages = sio.loadmat(filename)
    lP = labelsPassages['labelsPassageForEachSentence']
    return lP.flatten()

# this might be data-dependent
def load_passage_categories(filename):
    """
    Given .mat file, load and return list of passage category labels
    """
    labelsPassages = sio.loadmat(filename)
    lP = labelsPassages['keyPassageCategory']
    return list(np.hstack(lP[0]))

def get_passage_labels(f1g, lplst):
    """
    Given list of passage no. for each sentence, return list of passage no. (for each word)
    """
    lplst_word = []
    for i, sentence in enumerate(f1g):
        for word in sentence:
            lplst_word.append(lplst[i])
    return lplst_word

# this might be data-dependent
def load_passage_category(filename):
    """
    Given .mat file, return category no. for each passage
    """
    labelsPassageCategory = sio.loadmat(filename)
    lPC = labelsPassageCategory['labelsPassageCategory']
    lPC = np.hsplit(lPC,1)
    lpclst = np.array(lPC).tolist()
    lpclst = lpclst[0]
    lpclst = list(chain.from_iterable(lpclst)) # Accessing the nested lists
    return lpclst



def merge_lists(list_a, list_b, feature=""):
    '''Input: Two lists with potentially missing values.
       Return: If list 1 contains NA vals, the NA val is replaced by the value in list 2 (either numerical val or np.nan again)
    '''
    # count_a, count_b, count_na = 0, 0, 0
    merged = []

    for val1, val2 in zip(list_a, list_b):
        merged += [val2 if np.isnan(val1) else val1]

        # if not np.isnan(val1):
        #     merged.append(val1)
        #     count_b += 1
        # else:
        #     merged.append(val2)
        #     if not np.isnan(val2):
        #         count_a += 1
        #     else:
        #         count_na += 1
    
    return merged

    n = len(merged)
    print(feature, f"| number of values derived from original form: {count_b}, {count_b/n*100:.2f}%")
    print(feature, f"| number of values derived from lemmatized form: {count_a}, {count_a/n*100:.2f}%")
    print(feature, f"| number of values = NA: {count_na}, {count_na/n*100:.2f}%")
    print('-'*79)


# def compile_results(wordlst, wordlst_l, wordlst_lem, taglst, is_content_lst, setlst,
#                     snlst, lplst_word, snplst, wnslst, catlst, wordlen, merged_vals):
# def compile_results(wordlst, wordlst_l, wordlst_lem, 
#                     taglst, is_content_lst, setlst,
#                     snlst, wordlen, merged_vals):
#     """
#     Return dataframe: each row is a word & its various associated values
#     """
#     result = pd.DataFrame({'Word': wordlst})
#     result['Word cleaned'] = wordlst_l
#     result['Word lemma'] = wordlst_lem

#     result['POS'] = taglst
#     result['Content/function'] = is_content_lst
#     result['Set no.'] = setlst
#     result['Sentence no.'] = snlst
#     #result['Passage no.'] = lplst_word
#     #result['Sentence no. within passage'] = snplst
#     #result['Word no. within sentence'] = wnslst
#     #result['Broad topic'] = catlst
#     result['Specific topic'] = ['']*len(wordlst)
#     result['Word length'] = wordlen
#     result['polysemy'] = merged_vals['polysemy']

#     # List what you want the columns to be called
#     cols = {'NRC_Arousal': 'Arousal', 
#             'NRC_Valence': 'Valence', 
#             'OSC': 'Orthography-Semantics Consistency', 
#             'aoa': 'Age of acquisition', 
#             'concreteness': 'Concreteness',
#             'lexical_decision_RT': 'Lexical decision RT',
#             'log_contextual_diversity': 'Contextual diversity (log)',
#             'log_lexical_frequency': 'Lexical frequency (log)',
#             'n_orthographic_neighbors': 'Frequency of orthographic neighbors',
#             'num_morpheme': 'Number of morphemes',
#             'prevalence': 'Prevalence',
#             'surprisal-3': 'Lexical surprisal',
#             'total_degree_centrality': 'Degree centrality',
#             'polysemy':'Polysemy',
#             'num_morpheme_poly':'Number of morphemes poly',
#             #'Pronoun Ratio':'Pronoun Ratio'
#     }

#     for key, val in cols.items():
#         result[val] = merged_vals[key]
#     return result
    

    
# def conform_word_lex_df_columns(df):
#     # List what you want the columns to be called
#     cols = {'NRC_Arousal': 'Arousal', 
#             'NRC_Valence': 'Valence', 
#             'OSC': 'Orthography-Semantics Consistency', 
#             'aoa': 'Age of acquisition', 
#             'concreteness': 'Concreteness',
#             'lexical_decision_RT': 'Lexical decision RT',
#             'log_contextual_diversity': 'Contextual diversity (log)',
#             'log_lexical_frequency': 'Lexical frequency (log)',
#             'n_orthographic_neighbors': 'Frequency of orthographic neighbors',
#             'num_morpheme': 'Number of morphemes',
#             'prevalence': 'Prevalence',
#             'surprisal-3': 'Lexical surprisal',
#             'total_degree_centrality': 'Degree centrality',
#             'polysemy':'Polysemy',
#             'num_morpheme_poly':'Number of morphemes poly',
#     }
#     df.rename(columns=cols)

#     # Remove empty column that are vestiges of temporary analyses
#     df = df.drop(columns=['Specific topic'])
#     return df

# def transform_features(df, method='default', cols_log=None, cols_z=None):
#     df = df.copy()
#     if method == 'default':
#         cols_log = ['Degree centrality', 'Frequency of orthographic neighbors']
#     if cols_log:
#         for col in cols_log:
#             df[col] = np.log10(df[col].astype('float')+1)
#             df = df.rename({col: col+' (log)'})
#         df = df.rename(columns={col: col+' (log)' for col in cols_log})
#     if cols_z:
#         for col in cols_z:
#             df[col] = zscore(df[col].astype('float'), nan_policy='omit')
#         df = df.rename(columns={col: col+' (z)' for col in cols_z})
#     return df
    
# #     return df_main

# def countNA(lst):
#     """
#     Return number of NAs in a list
#     """
#     return sum(np.isnan(lst))

# def countNA_df(df, features='all'):
#     """
#     Given dataframe of words and feature values
#     Return list of number of NAs in each word's features
#     """
#     if features == 'all':
#         features = ['Age of acquisition', 'Concreteness', 'Prevalence', 'Arousal', 'Valence', 'Dominance', 'Ambiguity: percentage of dominant', 'Log lexical frequency', 'Lexical surprisal', 'Word length']
#     df = df[features]
#     return list(df.isnull().sum(axis=1))

# def uniqueNA(df, feature):
#     """
#     Given dataframe of words and feature values & desired feature,
#     return set of unique words with NA in given feature
#     """
#     return sorted(set(df['Word cleaned'][df[feature].isna()]))

# def avgNA(result, feature):
#     """
#     Return fractions of words with NA (for given feature) in each sentence
#     """
#     return result.groupby('Sentence no.').apply(lambda data: countNA(data[feature])/len(data))

# def get_NA_words(result, wordlst_l, features):
#     """
#     Return list of words that have NA in at least one of the specified features
#     """
#     big_u_lst = []
#     for feature in features:
#         big_u_lst.extend(uniqueNA(result, feature))
#     u_lst = sorted(set(big_u_lst))
#     return (big_u_lst, u_lst)




# def avg_feature(data, feature, method):
#     """
#     Return average value of feature
#     """
#     if method=='strict':
#         data = data.dropna()
#     elif method=='all':
#         pass
#     else:
#         raise ValueError('Method not recognized')
#     return np.nanmean(np.array(data[feature], dtype=float))

# def get_sent_vectors(df, features, method='strict', content_only=False,
#                      save=False, save_path=None, **kwargs):
#     """
#     Return dataframe of sentence embeddings (each row as a sentence)
#     Method:
#         'strict' - if a word has NA in any feature, it is skipped in the sentence average for all features
#         'all' - use all non-NA values for sentence average in any feature
#     content_only - if True, use content words only in a sentence
#     """
#     pronoun_ratios = kwargs.get('pronoun_ratios', None)
#     content_ratios = kwargs.get('content_ratios', None)
    
#     if content_only:
#         df = df[df["Content/function"] == 1]
#     sent_vectors = pd.DataFrame({'Sentence no.': df['Sentence no.'].unique()})
#     df = df[features + ['Sentence no.']].groupby('Sentence no.')
#     for name, feature in zip(features, features):
#         sent_vectors[name] = list(df.apply(lambda data: avg_feature(data, feature, method)))
#     if pronoun_ratios is not None:
#         sent_vectors['Pronoun ratios'] = pronoun_ratios['pronoun_ratio']
#     # if content_ratios is not None:
#     #     sent_vectors['Content ratios'] = content_ratios['content_ratio']
#     if save:
#         sio.savemat(save_path, {'sent_vectors': sent_vectors.drop(columns=['Sentence no.']).to_numpy()})
#     return sent_vectors

# def get_differential_sents(embed1, embed2, n, result, method='euclidean'):
#     """
#     Print sentences with the largest distance between the two input embeddings
#     Return index of these sentences (return 1-indexed; assume sentence no. are 1-indexed)
#     """
#     if method == 'euclidean':
#         func = ssd.euclidean
#     elif method == 'correlation':
#         func = ssd.correlation
#     elif method == 'cosine':
#         func = ssd.cosine
#     else:
#         raise ValueError('Method not implemented')
#     diff = np.array([func(embed1[i], embed2[i]) for i in range(len(embed1))])
#     top_diff_ind = (-diff).argsort(axis=None)[:n]
#     top_diff_sent_no = [i+1 for i in top_diff_ind]
#     print('Sentences with largest differences:', top_diff_sent_no)
#     for i, idx in enumerate(top_diff_ind):
#         sent_no = idx+1
#         sent = result[result['Sentence no.'] == sent_no].sort_values('Word no. within sentence')
#         print(f'{i+1}, sentence {sent_no}: ', list(sent['Word']))
#         # print('Number of NA features for a word:', countNA_df(sent, features='all'))
#         # print(f'Value in embedding 1: {x[idx]}, embedding 2: {y[idx]}')
#         print(f'Distance: {diff[idx]}')
#         print()
#     return top_diff_sent_no






########## PMI Block ##############
def GrabNGrams(sentences, save_paths):
    '''
    save paths is a list = 
        ['output_folder/03252021/PMI/example_pPMI_0.csv', 
        'output_folder/03252021/PMI/example_pPMI_1.csv', 
        'output_folder/03252021/PMI/example_pPMI_2.csv', 
        'output_folder/03252021/PMI/example_ngrams.pkl', 
        'output_folder/03252021/PMI/example_nm1grams.pkl']
    '''
    sample = sentences
    
    google1 = ZS('PMI/google-books-eng-us-all-20120701-1gram.zs')
    google2 = ZS('PMI/google-books-eng-us-all-20120701-2gram.zs')

    #  break sentences into strings
    def populate(sentences):
        ngra = dict()
        nm1gra = dict()
        for sentence in sentences:
            tokens = sentence.lower().split()
            tokens = ['_START_'] + tokens + ['_END_']
            for t in range(0, len(tokens) - 1):
                ngra[(tokens[t], tokens[t + 1])] = 0
                #print 0, (tokens[t], tokens[t + 1])
                nm1gra[tokens[t]] = 0
            for t in range(0, len(tokens) - 2):
                ngra[(tokens[t], tokens[t + 2])] = 0
                #print 1, (tokens[t], tokens[t + 2])
            for t in range(0, len(tokens) - 3):
                ngra[(tokens[t], tokens[t + 3])] = 0
                #print 2, (tokens[t], tokens[t + 3])
            nm1gra[tokens[len(tokens) - 1]] = 0
        for t1, t2 in ngra.copy().keys():
            ngra[(t2, t1)] = 0
        return ngra, nm1gra
    
    ngrams, nm1grams = populate(sample)
    
    #  fetch ngram and n-1gram
    def fetch(ngra, z=google2, zm1=google1):
        ngram_c = 0
        ngram_str = " ".join(ngra)
        #pdb.set_trace()
        for record in z.search(prefix=ngram_str):
            entry = record.split()
            if entry[1] == ngra[1]:
                ngram_c += int(entry[3])
        if nm1grams[ngra[0]] > 0:
            nm1gram_c = nm1grams[ngra[0]]
        else:
            nm1gram_c = 0
            for record in zm1.search(prefix=ngra[0]):
                entry = record.split()
                if entry[0] == ngra[0]:
                    nm1gram_c += int(entry[2])
        return ngram_c, nm1gram_c
    
    surprisals = dict()
    for ngram in ngrams.copy().keys():
        #print ngram
        #pdb.set_trace()
        ngrams[ngram], nm1grams[ngram[0]] = fetch(ngram)
    
    
    #with open(save_path+'/PMI/ngrams.pkl', 'w') as f:
    with open(save_paths[3], 'w') as f:
        pdb.set_trace()
        pickle.dump(ngrams, f)
    
    
    #with open('PMI/nm1grams.pkl', 'w') as f:
    with open(save_paths[4], 'w') as f:
        pickle.dump(nm1grams, f)


def pPMI(sentences, save_paths):
    '''
    save paths is a list = 
        ['output_folder/03252021/PMI/example_pPMI_0.csv', 
        'output_folder/03252021/PMI/example_pPMI_1.csv', 
        'output_folder/03252021/PMI/example_pPMI_2.csv', 
        'output_folder/03252021/PMI/example_ngrams.pkl', 
        'output_folder/03252021/PMI/example_nm1grams.pkl']
    '''
    sample = sentences
    
    with open('PMI/ngrams.pkl', 'r') as f:
        ngrams = pickle.load(f)
    
    
    with open('PMI/nm1grams.pkl', 'r') as f:
        nm1grams = pickle.load(f)
    
    N = 356033418959 # US american english v2 google ngrams
    nm1grams['_START_'] = float(sum([ ngrams[w] for w in ngrams.keys() if w[0] == '_START_']))
    
    
    def calc_prob(sentences, ngra=ngrams, nm1gra=nm1grams, ALPHA=0.1, lag=0):
        assert lag <= 2, 'impossible lag'
        results = []
        Z = len(ngrams.keys())*ALPHA + N
        for sent in sentences:
            string = sent[0]
            tokens = string.lower().split()
            mi = 0
            # No lag
            for t in range(0, len(tokens) - 1):
                joint_c = log(ngra[(tokens[t], tokens[t + 1])] + ngra[(tokens[t + 1], tokens[t])] + ALPHA)
                x_c = log(nm1gra[tokens[t]] + ALPHA * len(ngrams.keys()))
                y_c = log(nm1gra[tokens[t + 1]] + ALPHA * len(ngrams.keys()))
                pmi = max([0, (joint_c + log(Z) - x_c - y_c) / log(2)])
                mi += pmi
            # 1 word lag
            if lag >= 1:
                for t in range(0, len(tokens) - 2):
                    joint_c = log(ngra[(tokens[t], tokens[t + 2])] + ngra[(tokens[t + 2], tokens[t])] + ALPHA)
                    x_c = log(nm1gra[tokens[t]] + ALPHA * len(ngrams.keys()))
                    y_c = log(nm1gra[tokens[t + 2]] + ALPHA * len(ngrams.keys()))
                    pmi = max([0, (joint_c + log(Z) - x_c - y_c) / log(2)])
                    mi += pmi
            # 2 word lag
            if lag >= 2:
                for t in range(0, len(tokens) - 3):
                    joint_c = log(ngra[(tokens[t], tokens[t + 3])] + ngra[(tokens[t + 3], tokens[t])] + ALPHA)
                    x_c = log(nm1gra[tokens[t]] + ALPHA * len(ngrams.keys()))
                    y_c = log(nm1gra[tokens[t + 3]] + ALPHA * len(ngrams.keys()))
                    pmi = max([0,(joint_c + log(Z) - x_c - y_c) / log(2)])
                    mi += pmi
            results.append(','.join(sent[0].strip('\n'), str(mi)))
        return results
    
    
    result = calc_prob(sentences, lag=0)
    printstring = "\n".join(result)
    #with open('PMI/pPMI_0.csv', 'w') as f:
    with open(save_paths[0], 'w') as f:
        f.write(printstring)
    
    result = calc_prob(sentences, lag=1)
    printstring = "\n".join(result)
    #with open('PMI/pPMI_1.csv', 'w') as f:
    with open(save_paths[1], 'w') as f:
        f.write(printstring)
    
    result = calc_prob(sentences, lag=2)
    printstring = "\n".join(result)
    # with open('PMI/pPMI_2.csv', 'w') as f:
    with open(save_paths[2], 'w') as f:
        f.write(printstring)
########## End PMI Block

def sizeof_fmt(num, suffix='B'):
    '''
    This function can be used to print out how big a file is
    '''
    ''' by Fred Cirera,  https://stackoverflow.com/a/1094933/1870254, modified'''
    for unit in ['','Ki','Mi','Gi','Ti','Pi','Ei','Zi']:
        if abs(num) < 1024.0:
            return "%3.1f %s%s" % (num, unit, suffix)
        num /= 1024.0
    return "%.1f %s%s" % (num, 'Yi', suffix)




import os
import sentspace
from pathlib import Path

from nltk.tree import ParentedTree
from sentspace.syntax import utils
from sentspace.syntax.features import DLT, Feature, LeftCorner, Tree
from sentspace.utils import io, text
from sentspace.utils.caching import cache_to_disk

__pdoc__ = {'compute_tree_dlt_left_corner': False,
            'utils.calcEmbd': False,
            'utils.calcDLT': False,
            'utils.printlemmas': False,
            'utils.tree': False
            }

os.environ['PERL_BADLANG'] = '0'


def get_features(sentence: str = None, identifier: str = None, dlt: bool = True, left_corner: bool = True) -> dict:
    """executes the syntactic features pipeline

    Args:
        sentence (str, optional): exactly 1 sentence [None].
        dlt (bool, optional): calculate DLT feature? [False].
        left_corner (bool, optional): calculate Left Corner feature? [False].

    Returns:
        sentspace.syntax.features.Feature: a Feature instance with appropriate attributes
    """
    features = Feature()
    if dlt or left_corner:
        # io.log(f'parsing into syntax tree: `{sentence}`')
        # parsed = parse_input(sentence)
        features.tree = Tree(utils.compute_trees(sentence))
        # io.log(f'--- done: tree computed')
    else:
        return None

    # print(parse_input(sentence), features.tree)
    if dlt:
        # io.log(f'computing DLT feature')
        features.dlt = DLT(utils.compute_feature('dlt.sh', features.tree.raw), sentence, identifier)
        # io.log(f'--- done: DLT')
    if left_corner:
        # io.log(f'computing left corner feature')
        features.left_corner = LeftCorner(utils.compute_feature('leftcorner.sh', features.tree.raw), sentence, identifier)
        # io.log(f'--- done: left corner')

    tokenized = utils.tokenize(sentence).split()
    tagged_sentence = text.get_pos_tags(tokenized)
    is_content_word = utils.get_is_content(tagged_sentence, content_pos=text.pos_for_content)  # content or function word
    pronoun_ratio = utils.get_pronoun_ratio(tagged_sentence)
    content_ratio = utils.get_content_ratio(is_content_word)

    return {
        'index': identifier,
        'sentence': sentence,
        # 'tags': tagged_sentence,

        # 'content_words': is_content_word,
        'pronoun_ratio': pronoun_ratio,
        'content_ratio': content_ratio,

        # 'tree': features.tree
        'dlt': features.dlt, 
        'leftcorner': features.left_corner,
    }






from sentspace.utils.misc import *
download_nltk_resources()

import sentspace.utils.io as io
import sentspace.utils.caching as caching
import sentspace.utils.text as text
# import sentspace.utils.sanity_checks as sanity_checks

__pdoc__ = {
    "plotting": False,
    # "s3": False,
}



import os
from pathlib import Path
from typing import List

import sentspace.utils
from sentspace.lexical import utils
from sentspace.utils import io, text
from sentspace.utils.caching import cache_to_disk, cache_to_mem


def get_features(sentence: sentspace.Sentence.Sentence, lock=None) -> dict:

    # io.log(f'computing lexical featuures for `{sentence}`')

    # if lock: lock.acquire()
    databases = utils.load_databases(features='all')
    # if lock: lock.release()

    features_from_database = utils.get_all_features(sentence, databases)  # lexical features
    
    # TODO[?] return dict of lists, to be consistent with API?
    # return list of token-level features, as a dict per token
    returnable = []
    for i, token in enumerate(sentence.tokens):
        db_features_slice = {feature: features_from_database[feature][i] for feature in features_from_database}
        returnable += [{'index': sentence.uid,
                        'sentence': str(sentence),
                        'token': token,
                        'lemma': sentence.lemmas[i],
                        'tag': sentence.pos_tags[i],
                        'content_word': sentence.content_words[i],
                        **db_features_slice
                       }]

    return returnable




import re, os, sys, argparse
sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..', 'resource-gcg', 'scripts'))
import tree

argparser = argparse.ArgumentParser(description='''
Calculates Dependency Locality Theory integration cost values by word (along with some other useful fields).

Integration cost is implemented in 8 ways by varying both of the following binary parameters:

+/- CoordMod (c): Preceding conjuncts are skipped and only the weight of the heaviest conjunct is considered when calculating intervening discourse referents.
+/- VerbMod (v): Finite verbs weigh 2 EU (instead of 1) and non-finite verbs weigh 1 EU (instead of zero).
+/- ModMod (m): Preceding modifiers do not count as dependencies (and their distances are ommitted from the total integration cost).

Corresponding column names are "dlt(c)(v)(m)", with the letter present if the corresponding mod value is "+", and absent if the corresponding mod value is "-". For example, column "dltcm" is the integration cost using parameters +CoordMod, -VerbMod, +ModMod.

Input: Generalized Categorial Grammar (GCG15 or later) linetrees through stdin.
Output: A space-delimited data table with a row for each word in the input.
''')
argparser.add_argument('-d', '--debug', dest='DEBUG', action='store_true', help='Include debugging information about the calculation of dependencies in the output')
args, unknown = argparser.parse_known_args()

punc = ["-LRB-", "-RRB-", ",", "-", ";", ":", "\'", "\'\'", '\"', "`", "``", ".", "!", "?", "*FOOT*"]
phrasePunc = ["-RRB-", ",", "-", ";", ":", ".", "!", "?"]
pronouns = ['i', 'he', 'she', 's/he', 'it', 'we', 'they', 'you', 'me', 'him', 'her', 'us', 'them', \
        'who', 'whom', 'which', 'that', 'myself', 'himself', 'herself', 'itself', 'ourselves', 'themselves', 'oneself']
objExt = re.compile('.*-a.*-g.*')
terms = []
DLTcosts = []
DLTcostsV = []
coords = []
ends = []
complete = []
cCosts = []
cvCosts = []

def terminal(T):
    return len(T.ch) == 1 and len(T.ch[0].ch) == 0

def isPunc(t):
    if len(t.ch) == 0:
        if t.c not in punc:
            return False
        return True
    punctuation = True
    for c in t.ch:
        punctuation = punctuation & isPunc(c)
    return punctuation

def isPhrasePunc(t):
    if len(t.ch) == 0:
        if t.c not in phrasePunc:
            return False
        return True
    punctuation = True
    for c in t.ch:
        punctuation = punctuation & isPhrasePunc(c)
    return punctuation


def argsFromString(string):
    if string[-1] != '-':
        string += '-' #the hyphen triggers a push to the list, so this ensures last arg is pushed
    argMap = {}
    currentOp = 'head'
    if string[0:5] == '-LRB-' or string[0:5] == '-RRB-':
        argMap[currentOp] = [string[0:5]]
        string = string[5:]
        currentOp = ''
    currentArg = ''
    braceDepth = 0
    
    for char in string:
        if char == '-' and braceDepth == 0:
            if currentArg != '':
                if currentOp in argMap:
                    argMap[currentOp].append(stripBraces(currentArg))
                else:
                    argMap[currentOp] = [stripBraces(currentArg)]
            currentOp = ''
            currentArg = ''
        else:
            if currentOp == '':
                currentOp = char
            else:
                if char == '{':
                    braceDepth += 1
                if char == '}':
                    braceDepth -= 1
                currentArg += char
    return argMap

def stripBraces(string):
    if string[0] == '{':
        return string[1:-1]
    else:
        return string
        
def getPOS(T):
    if T.c[0] in punc:
        return 'null'
    if T.ch[0].c in pronouns:
        return 'pronoun' 
    if T.c == 'N-b{N-aD}':
        return 'D'
    if T.c[0] in ['B', 'G', 'L']:
        return 'Vnon'
    if T.c[0] in ['N', 'V', 'A', 'R', 'D', 'X']:
        return T.c[0]
    return 'O'
        
def head(T, dp=False):
    x = T
    while not terminal(x):
        if (len(x.ch)) == 1:
            x = x.ch[0]
        else:
            lArgs = argsFromString(x.ch[0].c)
            rArgs = argsFromString(x.ch[1].c)
            if 'l' not in lArgs:
                if x.ch[0].c == 'N-b{N-aD}' and not dp:
                    x = x.ch[1]
                else:
                    x = x.ch[0]
            elif 'l' not in rArgs:
                if lArgs['l'] == ['M'] and 'p' in rArgs and rArgs['p'] == ['Pc']:
                    x = x.ch[0]
                else:
                    x = x.ch[1]
            else:
                raise ValueError('-l rule specified on both children: ' + str(x.p))
    return x
    
def maxProj(T):
    x = T
    selfArgs = argsFromString(x.c)
    if x.p is not None and 'l' not in selfArgs and x.c != 'N-b{N-aD}' and (not x.c.startswith('X-cX-dX') or x.p.c.startswith('X-cX-dX')):
        return maxProj(x.p)
    else:
        return x

def first(T):
    x = T
    while not terminal(x):
        x = x.ch[0]
    return x

def last(T):
    x = T
    while not terminal(x):
        x = x.ch[-1]
    return x

def ancestor(T1, T2):
    x = T1
    while x.p != None:
        if x.p == T2:
            return True
        x = x.p
    return False

def objGap(t):
    if len(t.p.ch) < 2 or t != t.p.ch[1]:
        return False
    label = argsFromString(t.c)
    slabel = argsFromString(t.p.ch[0].c)
    if 'l' in slabel and 'N' in slabel['l'] \
            and 'a' in label and 'N' in label['a'] \
            and 'g' in label and 'N' in label['g']:
        return True


def calcCosts(t, MOD):
    tmp = t
    while t.p != None and len(t.p.ch) == 1:
        t = t.p
    s = getIdx(t) + 1
    dlt = 0
    dltc = 0
    dltcv = 0
    dltv = 0
    stop = isPunc(t) or t.c[0] == 'X' or t.p is None
    s = None
    depdirSyn = 'null'
    depdirSem = 'null'
    incr = False
    objGap = False
    while not stop:
        label = argsFromString(t.c)
        if not objGap:
            objGap = 'g' in label and 'N' in label['g'] and 'a' in label and 'N' in label['a']
        nonhead = ('l' in label and not (t.sibling() and t.sibling().c == 'N-b{N-aD}') and not 'C' in label['l']) or t.c.startswith('N-b{N-aD}')
        coord = ('l' in label and 'C' in label['l']) or ('c' in label) 
        if len(t.p.ch) == 2 and t == t.p.ch[1] and not isPunc(t.p.ch[0]) and not coord:
            slabel = argsFromString(t.p.ch[0].c)
            if MOD:
                if 'l' in slabel and 'M' in slabel['l']:
                    t = t.p
                    stop = nonhead or t.p is None
                    continue
            if 'l' in slabel and 'N' in slabel['l']:
                t = t.p
                stop = nonhead or t.p is None
                continue
            s = head(t.p.ch[0])
            e = tmp
            if 'l' in label:
                if 'M' in label['l']:
                    depdirSem = 'bwd'
                else:
                    depdirSem = 'fwd'
            else:
                depdirSem = 'bwd'
            if nonhead:
                if 'N' in label['l'] and objGap:
                    if args.DEBUG:
                        print('=====')
                        print('Object gap from "%s" to "%s"' % (e.ch[0].c, s.ch[0].c))
                        print('=====')
                    incr = True
                    depdirSyn = 'bwd'
                else:
                    depdirSyn = 'fwd'
            else:
                depdirSyn = 'bwd'
            if args.DEBUG:
                if incr:
                    wordlist = ' '.join([term.ch[0].c for term in terms[getIdx(s)+1:getIdx(e) + 1]])
                else:
                    wordlist = ' '.join([term.ch[0].c for term in terms[getIdx(s)+1:getIdx(e)]])
                if wordlist == '':
                    wordlist = '<EMPTY>'
                print('=====')
                print('Dependency with word "%s" in preceding subtree: %s' % (s.ch[0].c, t.p.ch[0]))
                print('Current word: ' + e.ch[0].c)
                print('List of intervening words: ' + wordlist)
                print('Dependency direction: ' + str(depdirSyn))
                print('=====')
            dlt += sumCosts(s, e, False, False, incr)
            dltc += sumCosts(s, e, False, True, incr)
            dltcv += sumCosts(s, e, True, True, incr)
            dltv += sumCosts(s, e, True, False, incr)
        t = t.p
        stop = nonhead or t.p is None
    return dlt, dltc, dltcv, dltv, depdirSyn, depdirSem
        
def getNext(t):
    tmp = t
    next = None
    while t.p != None and t == t.p.ch[-1]:
        t = t.p
    while not terminal(t):
        t = t.ch[-1]
    if t != tmp and terminal(t):
        next = t
    return next
    
def sumCosts(T1, T2, vMod, cMod, incr):
    s = getIdx(T1) + 1
    e = getIdx(T2)
    if incr:
        e += 1
    if cMod:
        if len(coords) > 0:
            for c in coords:
                if not ancestor(T1, c) and ancestor(T2, c):
                    return sumBetween(s, getIdx(first(c)), vMod, cMod) + sumBetween(getIdx(first(maxProj(T2))), e, vMod, cMod)
    if objExt.match(T2.c):
        e = getIdx(T2) + 1
    return sumBetween(s, e, vMod, cMod)
    
def sumBetween(s, e, vMod, cMod):
#    if cMod:
#        print DLTcosts[s:e]
#        print ' '.join([t.ch[0].c for t in terms[s:e]])
    sum = 0
    i = s
    if cMod:
        for j in range(0, len(complete)):
            cs = getIdx(first(complete[j]))
            ce = getIdx(last(complete[j]))
            if e < ce:
                break
            while i < cs:
                if vMod:
                    sum += DLTcostsV[i]
                else:
                    sum += DLTcosts[i]
                i += 1
            if i <= ce:
                i = ce + 1
                if vMod:
                    sum += cvCosts[j]
                else:
                    sum += cCosts[j]
        while i < e:
            if vMod:
                sum += DLTcostsV[i]
            else:
                sum += DLTcosts[i]
            i += 1
    else:
        if vMod:
            for i in range(s, e):
                sum += DLTcostsV[i]
        else:
            for i in range(s, e):
                sum += DLTcosts[i]
    return sum
    
def getIdx(T, terms_cur=None):
    if terms_cur is None:
        terms_cur = terms
    while not terminal(T) and len(T.ch) == 1:
        T = T.ch[0]
    if not terminal(T):
        raise ValueError('Cannot get index of non-terminal tree')
    return terms_cur.index(T)    
   
def getTerms(T):
    out = []
    if terminal(T):
        out.append(T)
    else:
        for x in T.ch:
            out += getTerms(x)
    return out

def cWeight(T, vMod):
    isCoord = False
    if terminal(T):
        if T.c[0] == 'N' and T.c != 'N-b{N-aD}':
            return 1
        if vMod:
            if T.c[0] == 'V':
                return 2
            if T.c[0] == 'G' or T.c[0] == 'L' or T.c[0] == 'B':
                return 1
        else:
            if T.c[0] == 'V':
                return 1
        return 0
    if len(T.ch) == 2:
        for t in T.ch:
            if "-c" in t.c:
                isCoord = True
    if isCoord:
        return max(cWeight(T.ch[0], vMod), cWeight(T.ch[1], vMod))
    weight = 0
    for t in T.ch:
        weight += cWeight(t, vMod)
    return weight
  
def followH(T, dp=False):
    label = argsFromString(T.c)
    out = None
    if 'h' in label:
        for x in T.ch:
            out = followH(x, dp=dp)
            if out is not None:
                break
    if out is None:
        out = head(T, dp=dp)
    return out

def followG(T, dp=False):
    label = argsFromString(T.c)
    out = None
    if 'g' in label:
        for x in T.ch:
            out = followH(x, dp=dp)
            if out is not None:
                break
    if out is None:
        out = head(T, dp=dp)
    return out

def followN(T, dp=False):
    label = argsFromString(T.c)
    out = None
    if 'r' in label and 'N' in label['r']:
        for x in T.ch:
            out = followH(x, dp=dp)
            if out is not None:
                break
    if out is None:
        out = head(T, dp=dp)
    return out

def getFwdDeps(T, terms_cur, dp=True):
    out = {}
    out_gap = {}
    if len(T.ch) == 2:
        deps = []
        gap_deps = []
        left = T.ch[0]
        right = T.ch[1]
        label_cur = argsFromString(T.c)
        label_left = argsFromString(left.c)
        label_right = argsFromString(right.c)
        if ('l' in label_left and 'A' in label_left['l']) \
                or ('l' in label_right and 'A' in label_right['l']) \
                or ('l' in label_left and 'M' in label_left['l']) \
                or ('l' in label_right and 'C' in label_right['l']):
            deps.append((head(left, dp=dp), head(right, dp=dp)))
        elif (('C' in label_cur['head'] or 'V' in label_cur['head'])) and ('r' in label_cur and 'N' in label_cur['r']) \
                and ('l' in label_cur and 'N' in label_cur['l']):
            deps.append((first(left), followG(right, dp=dp)))
            gap_deps.append((first(left), head(right, dp=dp)))
        elif 'i' in label_left and 'N' in label_left['i']:
            gap_deps.append((first(left), followG(right, dp=dp)))
        elif 'l' in label_right and 'N' in label_right['l'] and 'h' in label_left:
            gap_deps.append((followH(left), head(right, dp=dp)))
        
        for dep in deps:
            if not isPunc(dep[0]) and not isPunc(dep[1]):
                start = getIdx(dep[0], terms_cur)
                end = getIdx(dep[1], terms_cur)
                if end not in out:
                    out[end] = start

        for dep in gap_deps:
            if not isPunc(dep[0]) and not isPunc(dep[1]):
                start = getIdx(dep[0], terms_cur)
                end = getIdx(dep[1], terms_cur)
                if end not in out_gap:
                    out_gap[end] = start

    for x in T.ch:
        deps, gap_deps = getFwdDeps(x, terms_cur)
        
        for dep in deps:
            start = deps[dep]
            end = dep
            if end not in out:
                out[end] = start

        for dep in gap_deps:
            start = gap_deps[dep]
            end = dep
            if end not in out_gap:
                out_gap[end] = start

    return out, out_gap

def printToks(T, fwddeps=None, fwddeps_gap=None):
    top = fwddeps is None
    if top:
        terms_fwd = getTerms(T)
        fwddeps, fwddeps_gap = getFwdDeps(T, terms_fwd)
        # if args.DEBUG:
        if args.DEBUG:
            print('=====')
            print('Forward dependencies (for storage cost)')
            for x in fwddeps:
                print('%s -- %s' % (terms_fwd[fwddeps[x]].ch[0].c, terms_fwd[x].ch[0].c))
            for x in fwddeps_gap:
                print('%s -- %s' % (terms_fwd[fwddeps_gap[x]].ch[0].c, terms_fwd[x].ch[0].c))
            print('=====')
    if terminal(T):
        terms.append(T)
        discCost = 0
        discCostV = 0
        if T.c[0] == 'N' and T.c[0:9] != 'N-b{N-aD}' and T.ch[0].c.lower() not in pronouns:
            discCost = 1
            discCostV = 1
        elif T.c[0] == 'V' or T.c[0] == 'Q':
            discCost = 1
            discCostV = 2
        elif T.c[0] == 'G' or T.c[0] == 'L' or T.c[0] == 'B':
            discCostV = 1
        DLTcosts.append(discCost)
        DLTcostsV.append(discCostV)
       
        if args.DEBUG:
            print('=====')
            print('-ModMod Dependencies')
            print('=====') 
        ic, icc, iccv, icv, depdirSyn, depdirSem = calcCosts(T, False)
        dlt, dltc, dltcv, dltv = [discCost + ic, discCost + icc, discCostV + iccv, discCostV + icv]
        if args.DEBUG:
            print('=====')
            print('+ModMod Dependencies')
            print('=====')
        icm, iccm, iccvm, icvm, depdirSynM, depdirSemM = calcCosts(T, True)
        dltm, dltcm, dltcvm, dltvm = [discCost + icm, discCost + iccm, discCostV + iccvm, discCostV + icvm]
        if len(coords) > 0:
            if T == ends[-1]:
                while len(complete) > 0:
                    if ancestor(complete[-1], coords[-1]):
                        complete.pop()
                        cCosts.pop()
                        cvCosts.pop()
                    else:
                        break
                complete.append(coords[-1])
                cCosts.append(cWeight(complete[-1], False))
                cvCosts.append(cWeight(complete[-1], True))
                coords.pop()
                ends.pop()

        dlts = 0
        idx = getIdx(T)
        for e in fwddeps:
            if idx >= fwddeps[e] and idx < e:
                dlts += 1
        for e in fwddeps_gap:
            if idx >= fwddeps_gap[e] and idx < e:
                dlts += 1

        print(T.ch[0].c + ' ' + str(discCost) + ' ' + str(discCostV) + ' ' \
              + str(dlt) + ' ' + str(dltc) + ' ' + str(dltcv) + ' ' + str(dltv) + ' ' \
	          + str(dltm) + ' ' + str(dltcm) + ' ' + str(dltcvm) + ' ' + str(dltvm) + ' ' \
              + str(getPOS(T)) + ' ' + str(depdirSyn) + ' ' + str(depdirSem) + ' ' + str(depdirSynM) + ' ' + str(depdirSemM) + ' ' \
              + str(int(isPhrasePunc(T.ch[0]))) + ' ' + str(dlts)
            )
    else:
        if len(T.ch[0].c) > 1 and T.ch[0].c.endswith('lC') and not '-c' in T.c:
            coords.append(T)
            ends.append(last(T))
        for t in T.ch:
            printToks(t, fwddeps=fwddeps, fwddeps_gap=fwddeps_gap)

print('word dltdc dltdcv ' \
        + 'dlt dltc dltcv dltv ' \
        + 'dltm dltcm dltcvm dltvm ' \
        + 'pos depdirSyn depdirSem depdirSynM depdirSemM' + ' ' \
        + 'punc dlts')

for line in sys.stdin:
    if (line.strip() != '') and (line.strip()[0] != '%'):
        terms = []
        DLTcosts = []
        DLTcostsV = []
        coords = []
        ends = []
        complete = []
        cCosts = []
        cvCosts = []
        T = tree.Tree()
        T.read(line)
        printToks(T)



'''
'''

from collections import defaultdict
import typing
import sentspace.utils.text as text
import sentspace.utils.io as io
import re


class SentenceBatch(list):
    pass

class Sentence:
    '''
    a class to keep track of an individual sentence, its tokenized form,
    lemmas, POS tags, and so on.
    contains handy methods to perform the above segmentation/processing
    operations as well as methods for string representation, equality, and
    indexing (on tokenized form)
    '''
    _uid: str = None
    _raw: str = None
    _tokens: tuple = None
    _pos: tuple = None
    _lemmas: tuple = None
    _cleaned: tuple = None
    _content: tuple = None
    _lower: tuple = None
    
    def __init__(self, raw: str, uid: str = None, warn: bool = True) -> None:
        """Sentence constructor

        Args:
            raw (str): sentence in contiguous string form (will be segmented using a tokenizer)
            uid (str, optional): Unique ID of this sentence in a corpus. Defaults to None.
            warn (bool, optional): whether to warn that a UID is not supplied if one isn't given.
                                   can be set to False to suppress warning in case of intentional
                                   UID-less usage
        """        
        self._raw = re.sub(r' +', r' ', raw.strip())
        if uid is None and warn:
            io.log(f'no UID supplied for sentence {raw}\r', type='WARN')
        self._uid = uid
        self.OOV = defaultdict(set)

    def __hash__(self) -> int:
        return hash(self._raw)

    def __bool__(self) -> bool:
        '''Boolean value the sentence evaluates to.'''
        return self._raw != ''

    def __eq__(self, other) -> bool:
        """Equality operation with other sentences. Simply compares raw string.

        Args:
            other ([type]): [description]

        Returns:
            bool: True if self is equal to other, else False
        """        
        return str(self) == str(other)

    def __repr__(self) -> str:
        """Implement representation repr() of Sentence.
            prints out the UID followed by sentence

        Returns:
            str: representation of self
        """        
        return f'<{self._uid}>\t {self._raw:<32}'
    def __str__(self) -> str:
        """Returns raw string representation

        Returns:
            str: raw string
        """        
        return f'{self._raw}'

    def __len__(self) -> int:
        """Compute length of the sentence in terms of # of tokens

        Returns:
            int: # of tokens in this sentence (according to the default tokenization method `text.tokenize`)
        """        
        return len(self.tokens)

    def __getitem__(self, key: slice) -> str:
        """Support indexing into the tokens of the sentence

        Args:
            slice: slice of the tokenized sentence to retrieve

        Returns:
            str: token at the indexed position in the tokenized form
        """
        return self.tokens[key]

    def __iter__(self):
        '''we are iterable, so we return an iterator over tokens'''
        return iter(self.tokens)

    @property
    def uid(self) -> str:
        return self._uid

    @property
    def tokens(self) -> typing.Tuple[str]:
        return self.tokenized()
    # ^
    def tokenized(self, tokenize_method=text.tokenize) -> typing.Tuple[str]:
        """Tokenize and store tokenized form as a tuple. The tokenize_method is executed only
            the first time; each subsequent call to tokenized() returns the value of a stored variable

        Args:
            tokenize_method ([function], optional): a custom tokenization function. Defaults to utils.text.tokenize.

        Returns:
            tuple: tokenized form
        """        
        if self._tokens is None:
            self._tokens = tuple(tokenize_method(self._raw.lower()))
        return self._tokens

    # @property
    # def lowercased_tokens(self) -> typing.Tuple[str]:
    #     """Return lowercased tokens of this sentence

    #     Returns:
    #         typing.Tuple[str]: tuple of lowercased tokens from this sentence
    #     """        
    #     if self._lower is None:
    #         self._lower = [*map(lambda x: x.lower, self.tokenized())]
    #     return self._lower

    @property
    def pos_tags(self) -> typing.Tuple[str]:
        return self.pos_tagged()
    # ^
    def pos_tagged(self) -> typing.Tuple[str]:
        """POS-tag the sentence and return the result. Note, this method is also executed
            only once ever; each subsequent time a stored value is returned.

        Returns:
            tuple: POS tags
        """        
        if self._pos is None:
            self._pos = text.get_pos_tags(self.tokens)
        return self._pos

    @property
    def lemmas(self) -> typing.Tuple[str]:
        return self.lemmatized()
    # ^
    def lemmatized(self) -> typing.Tuple[str]:
        """Lemmatize and return the lemmas

        Returns:
            tuple: Lemmas
        """        
        if self._lemmas is None:
            self._lemmas = text.get_lemmatized_tokens(self.tokens, self.pos_tags)
        return self._lemmas

    @property
    def clean_tokens(self) -> typing.Tuple[str]:
        if self._cleaned is None:
            nonletters = text.get_nonletters(self.tokens, exceptions=[])
            self._cleaned = text.strip_words(self.tokens, method='punctuation', nonletters=nonletters)
        return self._cleaned

    @property
    def content_words(self) -> typing.List[int]:
        """Check whether each word is a 'content word'. Returns a list containing 0 or 1 as boolean
            mask indicating whether the word at that position is a content word or not

        Returns:
            List[int]: boolean mask indicating content words
        """        
        if self._content is None:
            self._content = text.get_is_content(self.pos_tags, content_pos=text.pos_for_content)
        return self._content



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sentspace',
 'sentspace.embedding',
 'sentspace.lexical',
 'sentspace.syntax',
 'sentspace.syntax.utils',
 'sentspace.utils']

package_data = \
{'': ['*'], 'sentspace': ['semantic/*', 'web/*']}

install_requires = \
['Morfessor>=2.0.6,<3.0.0',
 'PyICU>=2.8,<3.0',
 'PyYAML>=6.0,<7.0',
 'boto3>=1.20.25,<2.0.0',
 'nltk==3.6.2',
 'numpy>=1.22.1,<2.0.0',
 'pandas>=1.3.5,<2.0.0',
 'pdoc3>=0.10.0,<0.11.0',
 'polyglot>=16.7.4,<17.0.0',
 'pycld2>=0.41,<0.42',
 'scipy>=1.7.3,<2.0.0',
 'seaborn==0.11.2',
 'torch==1.9.0',
 'transformers==4.11.3']

setup_kwargs = {
    'name': 'sentspace',
    'version': '0.0.2',
    'description': '',
    'long_description': None,
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from collections import defaultdict
from contextlib import contextmanager
import os
import pickle
import typing
import warnings
from pathlib import Path
import random

import numpy as np
import pandas as pd
import sentspace.utils
from sentspace.utils import io, text
from sentspace.utils.caching import cache_to_mem #, cache_to_disk
from tqdm import tqdm

import torch
from transformers import AutoModel, AutoConfig, AutoTokenizer


def download_embeddings(model_name='glove.840B.300d.txt'):
    raise NotImplementedError
    if 'glove' in model_name:
        url = 'https://huggingface.co/stanfordnlp/glove/resolve/main/glove.840B.300d.zip'
        # download(url)




def flatten_activations(activations: typing.Dict[int, np.array]) -> pd.DataFrame:
    """
    Convert layer-wise activations into flattened dataframe format.
    Input: dict, key = layer, item = nd array of representations of that layer (n_tokens, )
    Output: pd dataframe, MultiIndex (layer, unit)
    """
    labels = []
    arr_flat = []
    for layer, act_arr in activations.items():
        arr_flat.append(act_arr.reshape(1,-1))
        for i in range(act_arr.shape[0]): # across units
            labels.append(('representation', layer, i,))
    arr_flat = np.concatenate(arr_flat, axis=1) # concatenated activations across layers
    df = pd.DataFrame(arr_flat)
    df.columns = pd.MultiIndex.from_tuples(labels) # rows: stimuli, columns: units
    return df


@cache_to_mem
def load_embeddings(emb_file: str = 'glove.840B.300d.txt',
                    data_dir: Path = None,
                    vocab: tuple = ()):
    """
    Read through the embedding file to find embeddings for target words in vocab
    Return dict mapping word to embedding (numpy array)
    """
    try:
        data_dir = Path(data_dir)
    except TypeError:
        data_dir = Path(__file__).parent / '..' / '..' / '.feature_database/'
    
    vocab = set(vocab)
    OOV = set(vocab)

    io.log(f"loading embeddings from {emb_file} for vocab of size {len(vocab)}")
    w2v = {}
    with (data_dir / emb_file).open('r') as f:
        total_lines = sum(1 for _ in tqdm(f, desc=f'counting # of lines in {data_dir/emb_file}'))
    with (data_dir / emb_file).open('r') as f:
        for line in tqdm(f, total=total_lines, desc=f'searching for embeddings in {emb_file}'):
            token, *emb = line.split(' ')
            if token in vocab or len(vocab) == 0:
                # print(f'found {token}!')
                w2v[token] = np.asarray(emb, dtype=float)
                OOV.difference_update({token}) # calling .remove() on an empty set would give an error
    
    io.log(f"---done--- loading embeddings from {emb_file}. OOV count: {len(OOV)}/{len(vocab)}")
    OOVlist = [*OOV]
    random.shuffle(OOVlist)
    io.log(f"           a selection of up to 32 random OOV tokens: {OOVlist[:32]}")

    return w2v


def get_word_embeds(sentence: sentspace.Sentence.Sentence, w2v: typing.Dict[str, np.array], 
                    model_name: str = 'glove', dims: int = None) -> typing.Dict[str, typing.DefaultDict[None, list]]:
    """Extracts [static] word embeddings for tokens in the given sentence 

    Args:
        sentence ([sentspace.Sentence.Sentence]): a Sentence object
        w2v ([dict]): word embeddings dictionary as a mapping from token -> vector
        model_name (str, optional): [description]. Defaults to 'glove'.
        dims (int, optional): [description]. Defaults to 300.

    Raises:
        ValueError: [description]

    Returns:
        [type]: [description]
    """
    # layer -> [emb_t1 emb_t2 emb_t3 ...]
    embeddings = defaultdict(list)
    dims = dims or next(iter(w2v[model_name].values())).shape[-1]

    # OOV_words = set()
    for token in sentence:
        if token in w2v:
            embeddings[0].append(w2v[token])
        else:
            embeddings[0].append(np.repeat(np.nan, dims))
            # sentence.OOV[model_name].add(token)
            # OOV_words.add(token)
    
    return {model_name: embeddings}


@cache_to_mem
def load_huggingface(model_name_or_path: str = 'distilgpt2', device='cpu'):
    """loads and caches a huggingface model and tokenizer

    Args:
        model_name_or_path (str): Huggingface model hub identifier or path to directory 
                                  containing config and model weights. Defaults to 'gpt2'.

    Returns:
        Tuple[AutoModel, AutoTokenizer]: returns a model and a tokenizer
    """
    if 'TRANSFORMERS_CACHE' not in os.environ:
        os.environ['TRANSFORMERS_CACHE'] = str(Path(__file__).parent.parent.parent / 'TRANSFORMERS_CACHE/')
    io.log(f"loading HuggingFace model [{model_name_or_path}] using TRANSFORMERS_CACHE={os.environ['TRANSFORMERS_CACHE']}")
    t = AutoTokenizer.from_pretrained(model_name_or_path, cache_dir=os.environ['TRANSFORMERS_CACHE'])
    m = AutoModel.from_pretrained(model_name_or_path, cache_dir=os.environ['TRANSFORMERS_CACHE'])
    m.to(device)
    m.eval()

    return m, t


def get_huggingface_embeds(sentence: sentspace.Sentence.Sentence, 
                           model_name: str = 'distilgpt2',
                           layers: typing.Collection[int] = None,
                           dims: int = None) -> typing.Dict[str, typing.DefaultDict[int, list]]:
    """Extracts [static] word embeddings for tokens in the given sentence 

    Args:
        sentence (sentspace.Sentence.Sentence): [description]
        model_name (str, optional): [description]. Defaults to 'distilgpt2'.
        layers (list[int], optional): a collection of layers to extract from the model. if None, all layers are extracted.
        dims (int, optional): [description]. Defaults to None.

    Raises:
        ValueError: [description]

    Returns:
        typing.Dict[str, typing.DefaultDict[int, list]]: [description]
    """
    # layer -> [emb_t1 emb_t2 emb_t3 ...]
    representations = dict()

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    model, tokenizer = load_huggingface(model_name, device=device)

    # we don't want to track gradients; only interested in the encoding
    model.eval()
    with torch.no_grad():

        # current procedure processes sentences individually. consider minibatching.
        batch_encoding = tokenizer(str(sentence), return_tensors="pt", truncation='longest_first').to(device)
        input_ids = batch_encoding['input_ids']
        
        # overflow_tokens = max(0, len(input_ids) - model.config.n_positions)
        # if overflow_tokens > 0: io.log(f"Stimulus too long! Truncated the first {overflow_tokens} tokens", type='WARN')
        
        # input_ids = input_ids[overflow_tokens:]
        # print(tokenizer.convert_ids_to_tokens(input_ids))

        output = model(input_ids, output_hidden_states=True, return_dict=True)
        hidden_states = output['hidden_states']

        #  for i in range(n_layer+1):
        for layer in range(len(hidden_states)):
            if layers is None or layer in layers:
                token = slice(None, None) # placeholder to allow a possibility of picking a particular token rather than the full sequence
                representations[layer] = hidden_states[layer].detach().cpu().squeeze().numpy()[token, :]

        # print(input_ids.shape, representations[0].shape)

    return {model_name: representations}

    

def pool_sentence_embeds(sentence, token_embeddings, filters={'nofilter': lambda i, x: True},
                         keys=None, methods={'mean', 'median'}):
    """pools embeddings of an entire sentence (given as a list of embeddings)
       using averaging, maxpooling, minpooling, etc., after applying all the
       provided filters as functions (such as content words only).

    Args:
        token_embeddings (list[np.array]): [description]
        filters (list[function[(idx, token) -> bool]], optional): [description]. Defaults to [lambda x: True].
            filters should be functions that map token to bool (e.g. is_content_word(...))
            only tokens that satisfy all filters are retained.
        keys (`typing.Union[typing.Collection[str], None]`): which models we want to pool using the methods supplied. 
            if None, all available models are pooled (separately) using the supplied methods
        methods (`typing.Collection[typing.Union[str, typing.Tuple[str, typing.Callable]]]`):

    Returns:
        dict: averaging method -> averaged embedding
    """                         

    """
    Return dataframe of each sentence no. and its sentence embedding
    from averaging embeddings of words in a sentence (ignore NAs)
    Parameters:
        df: dataframe, output of get_glove_word()
        content_only: if True, use content words only
        is_content_lst: list, values 1 if token is content word, 0 otherwise
        save: whether to save results
        save_path: path to save, support .csv & .mat files
    """
    
    # model -> method -> repr
    all_pooled = defaultdict(dict)

    for model_name in token_embeddings:

        # if the current model_name is not meant to be aggregated in this manner, skip
        # (e.g. BERT and last token "aggregation")
        # if keys is None, any aggregation step specified will be applied regardless of model_name
        # (e.g., mean)
        if keys and model_name not in keys: 
            continue

        # map from method --> layer (int) --> pooled representation per model
        # this entity still needs to be flattened
        model_pooled = defaultdict(dict)

        for layer in token_embeddings[model_name]:

            # all the embeddings corresponding to the tokens
            all_embeds = [e for i, (t, e) in enumerate(zip(sentence.tokens, token_embeddings[model_name][layer]))]
            all_tokens = [t for i, (t, e) in enumerate(zip(sentence.tokens, token_embeddings[model_name][layer]))]
            all_embeds = np.array(all_embeds, dtype=np.float32)
            all_tokens = np.array(all_tokens, dtype=str)
            # print(all_tokens, all_embeds.shape)

            # exclude OOV words' embeddings (they are all NaNs)
            not_nan_tokens = all_tokens[~np.isnan(all_embeds[:, 0])   ]
            not_nan_embeds = all_embeds[~np.isnan(all_embeds[:, 0]), :]

            # make a note of the shape of the vector (n x embed_dim)
            shape = not_nan_embeds.shape
            # TODO: vectorize operation on all tokensnow that it is an numpy array using apply()?
            if filters:
                mask = [all(fn(i, t) for fn_name, fn in filters.items()) for i, t in enumerate(not_nan_tokens)]
            else:
                mask = slice(None, None, None)

            filtered_embeds = not_nan_embeds[mask]
            filtered_shape = filtered_embeds.shape

            # if filtering left no tokens, we will use all_embeds instead
            if filtered_shape[0] == 0:
                io.log(f'filtered embeddings for current sentence are empty. retrying without filters: {sentence.tokens}', type='WARN')

                # now what? use unfiltered (as a fallback)
                filtered_embeds = not_nan_embeds
                
            # [very rarely] if no word has a corresponding embedding, then we have no choice
            # but to return a zero vector (or, sometime in the future, a random vector??)
            if shape[0] == 0:
                filtered_embeds = np.zeros((1, shape[-1]))

            for method in methods:
                # if a pre-defined aggregation method is used, apply it
                if type(method) is str:
                    method_name = method
                    if method_name == 'median':
                        pooled = np.median(filtered_embeds, axis=0).reshape(-1)#.tolist()
                    elif method_name == 'mean':
                        pooled = filtered_embeds.mean(axis=0).reshape(-1) #.tolist()
                    elif method_name == 'last':
                        pooled = filtered_embeds[-1, :].reshape(-1) #.tolist()
                    elif method_name == 'first':
                        pooled = filtered_embeds[0, :].reshape(-1) #.tolist()
                    else:
                        raise ValueError(f'unknown pooling method identifier: {method}')
                # handle the case where a custom aggregation function is applied to the embeddings
                elif type(method) is tuple:
                    method_name, fn = method
                    pooled = fn(filtered_embeds).reshape(-1) #.tolist()

                else:
                    raise ValueError(method)

                model_pooled[method_name][layer] = pooled

        for method_name, layer_wise_reprs in model_pooled.items():
            all_pooled[model_name][method_name] = flatten_activations(layer_wise_reprs)
            all_pooled[model_name][method_name].index = [sentence.uid]

    return all_pooled




import sys, re, argparse
sys.setrecursionlimit(1500)

import pdb

## Command-line processing
argparser = argparse.ArgumentParser('''
Calculates various center-embedding fields on %.nopunc.tss.fromlinetree.semprocdecpars and outputs tokmeasures using tokenization provided by %.linetoks (stdin).
''')
argparser.add_argument('decpars', nargs=1, help='%.nopunc.tss.fromlinetree.semprocdecpars file to use for embedding calculations.')
argparser.add_argument('-d', '--debug', action='store_true', help='Print verbose debugging log instead of data table.')
args, unknown = argparser.parse_known_args()

## Global variables
i = 0
pronouns = ['I', 'he', 'she', 's/he', 'it', 'we', 'they', 'you', 'me', 'him', 'her', 'us', 'them', \
        'who', 'whom', 'which', 'that', 'myself', 'himself', 'herself', 'itself', 'ourselves', 'themselves', 'oneself']
depth_modes = ['any', 'allnolo', 'cc', 'min']
length_measures = ['len', 'dr', 'drv']

## Regex searches
s = re.compile('----(.*)$')
w = re.compile('W .* ([^ ]+)$')
f = re.compile('F .* ([01])[^ ]*$')
j = re.compile('J .* ([01])[^ ]*$')
p = re.compile('P .* ([^ ]+)$')
frag = re.compile(' *\[[^\]]*\]:([^\/;]+)\/ *\[[^\]]*\]:([^\/;]+)')
#frag = re.compile(' *[^\]]+\]:([^\/]+)\/*[^\]]+\]:([^\/]+)')
r = re.compile('====(.*)$')
maxsplit = re.compile('[/;]')

## Class definitions
class StoreNode(object):
    def __init__(self, t, f, j, lab, edgetype='A', dr=0, drv=0, prev=None, next=None, above=None, below=None):
        self.t = t
        self.f = f
        self.j = j
        self.lab = lab
        self.edgetype = edgetype
        self.dr = dr
        self.drv = drv
        self.prev = prev
        self.next = next
        self.above = above
        self.below = below
        self.measures = ['len', 'dr', 'drv']

        self.d = {}
        for mode in depth_modes: 
            self.d[mode] = self.depth(mode)
    
    def __str__(self):
        out = self.lab if self.lab != None else '<EMPTY>'
        if self.above != None:
            out = str(self.above) + ' <--' + self.edgetype + ' ' + out
        else:
            out = '%s F:%s J:%s DR:%s DRV:%s ' %(self.t, self.f, self.j, self.dr, self.drv) + out
            if self.prev != None:
                out = str(self.prev.bottom()) + '\n' + out
        return out

    def retrievalLen(self, mode, measure, definition):
        n = self
        noF = n.f == 0
        endembd = n.endembd(mode) == 1
        length = 0
        
        if definition == 'noF':
            if not noF:
                return 0
        elif definition == 'embd':
            if not endembd:
                return 0
        elif definition == 'reinst':
            if n.f == 1 and n.endembd(mode) == 0:
                return 0
        else:
            raise ValueError('Length definition "%s" not supported.' %definition)

        if measure == 'len':
            length += 1
        elif measure == 'dr':
            length += n.dr
        elif measure == 'drv':
            length += n.drv
        else:
            raise ValueError('Length measure type "%s" not supported' %measure)
        
        d = n.d[mode]
        prev = n.top().prev
        if prev == None:
            return length

        n = prev
        if endembd:
            d += 1
        while n.d[mode] < d and n.below != None:
            n = n.below
        while n.d[mode] > d and n.above != None:
            n = n.above
        prev = n.top().prev
        
        stop = n.d[mode] != d or n == None
        if not stop and (definition == 'noF' or (definition=='reinst' and not endembd)):
            stop = n.f==0 and not n.endembd(mode)==1
        while not stop:
            if measure == 'len':
                length += 1
            elif measure == 'dr':
                length += n.dr
            elif measure == 'drv':
                length += n.drv
            n = prev
            if n != None:
                prev = n.top().prev
                while n.d[mode] < d and n.below != None:
                    n = n.below
                while n.d[mode] > d and n.above != None:
                    n = n.above
            stop = n == None or n.d[mode] != d
            if not stop and (definition == 'noF' or (definition=='reinst' and not endembd)):
                stop = n.f==0 and not n.endembd(mode)==1
        return length
            

    def depth(self, mode='any'):
        d = 0
        n = self
        if self.below != None:
            belowedgetype = self.below.edgetype
        else:
            belowedgetype = 'B'
        while n != None:
            if mode == 'any':
                d += 1
            elif mode == 'allnolo':
                if n.above != None and \
                     ((n.edgetype == 'A' and belowedgetype == 'B') \
                   or (n.edgetype == 'A' and belowedgetype == 'A') \
                   or (n.edgetype == 'B' and belowedgetype == 'B')):
                    d += 1
            elif mode == 'cc':
                if n.edgetype == 'A' and n.above != None:
                    d += 1
            elif mode == 'Bonly':
                if n.edgetype == 'B' and n.above != None:
                    d += 1
            elif mode == 'min':
                if n.edgetype == 'A' and belowedgetype == 'B':
                    d += 1
            else:
                raise ValueError('Mode "%s" for depth calculation is not supported' %mode)
            belowedgetype = n.edgetype
            n = n.above
        return d

    def get_ab(self):
        ad = None
        bd = None
        n = self
        if self.below != None:
            belowedgetype = self.below.edgetype
        else:
            belowedgetype = 'B'
        while n != None and (ad == None or bd == None):
            if belowedgetype == 'A' and n.edgetype == 'B' or n.above == None:
                ad = n.lab
            elif belowedgetype == 'B' and n.edgetype == 'A' or n.above == None:
                bd = n.lab
            belowedgetype = n.edgetype
            n = n.above
        adm1 = None
        bdm1 = None
        while n != None and (adm1 == None or bdm1 == None):
            if belowedgetype == 'A' and n.edgetype == 'B' or n.above == None:
                adm1 = n.lab
            elif belowedgetype == 'B' and n.edgetype == 'A':
                bdm1 = n.lab
            belowedgetype = n.edgetype
            n = n.above
         
        return ad if ad != None else 'null', \
               bd if bd != None else 'null', \
               adm1 if adm1 != None else 'null', \
               bdm1 if bdm1 != None else 'null'
                
    def top(self):
        n = self
        while n.above != None:
            n = n.above
        return n

    def bottom(self):
        n = self
        while n.below != None:
            n = n.below
        return n

    def endembd(self, mode):
        dt = self.d[mode]
        if self.prev != None:
            dtm1 = self.prev.bottom().d[mode]
            out = int(dt < dtm1)
        else:
            out = 0
        return out

    def startembd(self, mode):
        dt = self.d[mode]
        if self.prev != None:
            dtm1 = self.prev.bottom().d[mode]
            out = int(dt > dtm1)
        else:
            out = 1
        return out


## Method definitions
def next_tok(buffer):
    line = decpars.readline()
    while line and not line.startswith('N '):
        line = decpars.readline()
    return(line)

def process_tok(buffer, line, i, prev, debug=True):
    while line and not line.startswith('F '):
        line = buffer.readline()
    assert line, 'Badly-formed input'
    F = int(f.match(line.strip()).group(1))
    while line and not line.startswith('P '):
        line = buffer.readline()
    assert line, 'Badly-formed input'
    pos = p.match(line.strip()).group(1)
    while line and not line.startswith('W '):
        line = buffer.readline()
    assert line, 'Badly-formed input'
    word = w.match(line.strip()).group(1)
    DR = dr(pos, word)
    DRV = drv(pos, word)
    while line and not line.startswith('J '):
       line = buffer.readline()
    assert line, 'Badly-formed input'
    J = int(j.match(line.strip()).group(1))
    while line and not (line.startswith('----') or line.startswith('TREE ')):
       line = buffer.readline()
    if line and line.startswith('----'):
        storestr = s.match(line.strip()).group(1)
        node = parse_store(storestr, i, F, J, prev, DR, DRV)
        i += 1
        prev = node.top()
    else:
        node = StoreNode(i, F, J, None, prev=prev, dr=DR, drv=DRV)
        i = 0
        prev = None
    
    preds = [] 
    preds += [word, DR, DRV, 1-F, J]

    embddepthAny, embddepthAllnolo, embddepthCC, embddepthMin = [node.d[mode] for mode in depth_modes]
    preds += [embddepthAny, embddepthAllnolo, embddepthCC, embddepthMin]
    
    endembdAny, endembdAllnolo, endembdCC, endembdMin = [node.endembd(mode) for mode in depth_modes]
    preds += [endembdAny, endembdAllnolo, endembdCC, endembdMin]
    
    startembdAny, startenmbdAllnolo, startembdCC, startembdMin = [node.startembd(mode) for mode in depth_modes]
    preds += [startembdAny, startenmbdAllnolo, startembdCC, startembdMin]
    
    noFlen, noFdr, noFdrv = [node.retrievalLen('min', measure, 'noF') for measure in length_measures]
    preds += [noFlen, noFdr, noFdrv]
    
    embdlen, embddr, embddrv = [node.retrievalLen('min', measure, 'embd') for measure in length_measures]
    preds += [embdlen, embddr, embddrv]
    
    reinstlen, reinstdr, reinstdrv = [node.retrievalLen('min', measure, 'reinst') for measure in length_measures]
    preds += [reinstlen, reinstdr, reinstdrv]
    
    Ad, Bd, Adm1, Bdm1 = node.get_ab()
    AdPrim = get_primitive_cat(Ad)
    BdPrim = get_primitive_cat(Bd)
    Bdm1Prim = get_primitive_cat(Bdm1)
    Adm1Prim = get_primitive_cat(Adm1)
    preds += [Ad, AdPrim, Bd, BdPrim, Bdm1, Bdm1Prim]
    
    if debug:
        print('-'*50)
        print(word)
        print(node)
        print('embddepthAny embddepthAllnolo embddepthCC embddepthMin')
        print('%s %s %s %s' %(embddepthAny, embddepthAllnolo, embddepthCC, embddepthMin))
        print('endembdAny endembdAllnolo endembdCC endembdMin')
        print('%s %s %s %s' %(endembdAny, endembdAllnolo, endembdCC, endembdMin))
        print('startembdAny startenmbdAllnolo startembdCC startembdMin')
        print('%s %s %s %s' %(startembdAny, startenmbdAllnolo, startembdCC, startembdMin))
        print('noFlen noFdr noFdrv')
        print('%s %s %s' %(noFlen, noFdr, noFdrv))
        print('embdlen embddr embddrv') 
        print('%s %s %s' %(embdlen, embddr, embddrv))
        print('reinstlen reinstdr reinstdrv')
        print('%s %s %s' %(reinstlen, reinstdr, reinstdrv))
        print('Ad AdPrim Bd BdPrim Adm1 Adm1Prim Bdm1 Bdm1Prim')
        print('%s %s %s %s %s %s %s %s' %(Ad, AdPrim, Bd, BdPrim, Adm1, Adm1Prim, Bdm1, Bdm1Prim))
        print('-'*50)
            
    return i, prev, preds 

def parse_store(storestr, t, f, j, prev=None, DR=0, DRV=0):
    cur = ''
    edgetype = 'A'
    lab = None
    above = None
    node = None
    while storestr != '':
        char = storestr[0]
        storestr = storestr[1:]
        if char == '/' or char == ';':
            lab = cur.split(':')[-1]
            node = StoreNode(t, f, j, lab, edgetype=edgetype, dr=DR, drv=DRV, above=above, prev=prev)
            if above != None:
                above.below = node
            if prev != None:
                prev = prev.below
            above = node
            cur = ''
            if char == ';':
                edgetype = 'B'
            else:
                edgetype = 'A'
        else:
            cur += char

    return node
    
def dr(pos, word):
    if pos.startswith('V') or ((pos.startswith('N') and not pos.startswith('N-b{N-aD}')) and not word in pronouns):
        return 1
    else:
        return 0

def drv(pos, word):
    if pos.startswith('V'):
        return 2
    elif pos[0] in ['V','G','B','L'] or ((pos.startswith('N') and not pos.startswith('N-b{N-aD}')) and not word in pronouns):
        return 1
    else:
        return 0

def get_primitive_cat(s):
    if s == 'N-b{N-aD}':
        return 'D'
    if s[0] in ['B', 'G', 'L']:
        return 'Vnon'
    if s[0] in ['N', 'V', 'A', 'R', 'D', 'X', 'S']:
        return s[0]
    if s == 'null':
        return s
    return 'O'




## Main program

header = []
punc_tok = []
header += 'word', 'dr', 'drv'
punc_tok += ['0'] * 2
header += 'noF', 'yesJ'
punc_tok += ['0'] * 2
header += 'embddepthAny', 'embddepthAllnolo', 'embddepthCC', 'embddepthMin'
punc_tok += ['0'] * 4
header += 'endembdAny', 'endembdAllnolo', 'endembdCC', 'endembdMin'
punc_tok += ['0'] * 4
header += 'startembdAny', 'startenmbdAllnolo', 'startembdCC', 'startembdMin'
punc_tok += ['0'] * 4
header += 'noFlen', 'noFdr', 'noFdrv'
punc_tok += ['0'] * 3
header += 'embdlen', 'embddr', 'embddrv'
punc_tok += ['0'] * 3
header += 'reinstlen', 'reinstdr', 'reinstdrv'
punc_tok += ['0'] * 3
header += 'Ad', 'AdPrim', 'Bd', 'BdPri', 'Bdm1', 'Bdm1Prim'
punc_tok += ['null'] * 6
print(' '.join(header))
    

with open(args.decpars[0], 'rb') as decpars:
    dpars_line = next_tok(decpars)
    ltoks_line = sys.stdin.readline()
    prev = None
    i, prev, preds = process_tok(decpars, dpars_line, i, prev, debug=args.debug)
    while ltoks_line and dpars_line:
        for wrd in ltoks_line.strip().split():
            if preds is not None and wrd == preds[0]:
                print(' '.join([str(P) for P in preds]))
                if dpars_line:
                    i, prev, preds = process_tok(decpars, dpars_line, i, prev, debug=args.debug)
                    dpars_line = next_tok(decpars)
                else:
                    preds = None
            else: ## punctuation token
                print(' '.join([wrd] + punc_tok))
        ltoks_line = sys.stdin.readline()
    if preds is not None:
        print(' '.join([str(P) for P in preds]))






import typing
from joblib import Memory
from pathlib import Path
from functools import lru_cache, partial


_cache_instance = Memory(location=Path(__file__).parent / '..' / '.cached_function_calls', verbose=0, mmap_mode='c')

def cache_to_disk(fn):
    '''
    Decorator that enables caching function calls to the disk at 
    `Path(__file__).parent / '..' / '.cached_function_calls'`.
    E.g.

        @cache_to_disk
        def fibonacci(n):
            ...
    '''
    return _cache_instance.cache(fn)


def cache_to_mem(fn):
    '''
    Decorator that acts as an alias of `lru_cache(maxsize=None)`
    E.g.

        @cache_to_mem
        def fibonacci(n):
            ...
    '''
    @lru_cache(maxsize=None)
    def wrapped_fn(*args, **kwargs): 
        return fn(*args, **kwargs)

    return wrapped_fn

# DEBUG: uncomment below line to disable caching by making a dummy wrapper
# cache_to_disk = lambda x: x





import string
from collections import Counter

import pandas as pd
import sentspace.utils
from nltk import pos_tag
from nltk.corpus import wordnet
from nltk.stem import WordNetLemmatizer
from sentspace.utils.caching import cache_to_disk, cache_to_mem

from nltk.tokenize import TreebankWordTokenizer
tokenize = TreebankWordTokenizer().tokenize

pos_for_lemmatization = (wordnet.ADJ, wordnet.VERB, wordnet.NOUN, wordnet.ADV) # define POS used for lemmatization
pos_for_content = (wordnet.ADJ, wordnet.VERB, wordnet.NOUN, wordnet.ADV)  # define POS that count as content words

# @cache_to_mem
def get_pos_tags(sentence:tuple) -> tuple:
    """
    Given sentence (a list of tokens), return single list of POS tags
    """
    return tuple(tag for token, tag in pos_tag(sentence))


def get_flat_pos_tags(token_lists):
    """
    Given list of sentences (each a list of tokens), return single list of POS tags
    """
    sentspace.utils.io.log('`get_flat_post_tags` '
        'applied to multiple sentences will be DEPRECATED. '
        'please call the appropriate function corresponding to a single sentence.', type='WARN')

    all_pos_tags = []
    for sentence_tokens in token_lists:
        tags = get_pos_tags(sentence_tokens)
        for tag in tags:
            all_pos_tags.append(tag[1])
    return all_pos_tags


def get_flat_sentence_num(token_lists):
    """returns a flattened sentence number list from a list of sentences 

    Args:
        token_lists (List[List[str]]): list of sentences, each of which is a list of tokens

    Returns:
        List[str]: flattened list of sentence number corresponding to each token
    """
    return [1+i for i, sentence in enumerate(token_lists) for token in sentence]


def get_flat_tokens(token_lists):
    '''returns a flattened sentence number list from a list of sentences 

    Args:
        token_lists (List[List[str]]): list of sentences, each of which is a list of tokens

    Returns:
        List[str]: flattened list of tokens
    '''
    return [token for sentence in token_lists for token in sentence]


# def get_sent_num(token_lists):
#     """
#      Given list of sentences (each a list of tokens),
#      return list of sentence no. for each token (starting at 1)
#     """
#     sentence_numbers = []  # sentence number list
#     for i, sentence in enumerate(token_lists):
#         for word in sentence:
#             snlst.append(i+1)
#     return snlst


def get_flat_word_num(token_lists):
    """
    Given list of sentences (each a list of tokens), return list of word no.
    """
    return [1+i for sentence in token_lists for i, token in enumerate(sentence)]


def get_passage_category(lpclst, lplst_word, keyPassageCategory):
    """
    Given category no. for each passage & list of string labels (corresponding to category no.),
    Returns
        1. list of passage category labels for each word
        2. dict mapping category no. to string label
    """
    cat_dict = {num+1: cat for num, cat in enumerate(
        keyPassageCategory)}  # maps category number to category string
    catlst = []
    for passage_num in lplst_word:
        cat_num = lpclst[passage_num-1]
        catlst.append(cat_dict[cat_num])
    return catlst, cat_dict


def get_nonletters(wordlst:tuple, exceptions={'-'}) -> set:
    """
    Given list of tokens, print & return all non-alphabetical characters (except for input exceptions)
    For helping configure how to clean words
    """
    all_chars = set()
    all_nonletters = set()

    for word in wordlst:
        chars, nonletters = get_nonletters_from_word(word, exceptions)
        all_chars.update(chars)
        all_nonletters.update(nonletters)

    # print('All unique characters:', sorted(charlst))
    # print('All non-letter characters in text file:', sorted(nonletters))
    # print('Exceptions passed in:', sorted(exceptions))
    # print('-'*79)
    return all_nonletters

# @cache_to_disk
def get_nonletters_from_word(word, exceptions):
    charlst = set()
    nonletters = set()
    for char in word:
        charlst.add(char)
        if not char.isalpha() and char not in exceptions:
            nonletters.add(char)
    return charlst, nonletters


def strip_words(flat_token_list, method='nonletter', nonletters=None, punctuation=string.punctuation):
    """
    Given list of tokens, return list of cleaned/stripped words. Lower-cased.
    method = 'nonletter': remove all nonletters specified in the nonletters argument
    method = 'punctuation': remove all nonletters specified in the punctuation argument (default value from string.punctuation)
    """
    # if method == 'nonletter':
    #     print('Formatting words - Characters ignored:', nonletters)
    # elif method == 'punctuation':
    #     print('Formatting words - Characters ignored:', punctuation)
    # print('-'*79)
    flat_cleaned_token_list = []
    for t in flat_token_list:
        stripped = strip_word(t, method, nonletters, punctuation)
        if stripped: flat_cleaned_token_list.append(stripped.lower())
    return flat_cleaned_token_list


# @cache_to_disk
def strip_word(word:str, method:str, nonletters:set, punctuation:set) -> str:
    stripped = ''
    for char in word:
        if method == 'nonletter':
            if char not in nonletters:
                stripped += char
        elif method == 'punctuation':
            if char not in punctuation:
                stripped += char
        else:
            raise ValueError(f'unknown method passed to strip word {method}')
    return stripped


def get_token_lens(flat_token_list):
    """get the length of each token in list

    Args:
        flat_tokens_list (List[str]): list of tokens

    Returns:
        List[int]: list of respective lengths
    """    
    return [*map(len, flat_token_list)]


def get_lemmatized_tokens(flat_token_list, flat_pos_tags, lemmatized_pos=[wordnet.ADJ, wordnet.VERB, wordnet.NOUN, wordnet.ADV]):
    """
    Given list of tokens & tags, lemmatize nouns & verbs (as specified by input POS tags)
    """
    
    lemmas = WordNetLemmatizer()
    @cache_to_mem
    def lemmatize_token(word, pos):
        if pos in lemmatized_pos:
            return lemmas.lemmatize(word, pos)
        return word

    lemmatized_tokens = []
    for word, POS in zip(flat_token_list, flat_pos_tags):
        pos = get_wordnet_pos(POS)
        lemmatized_tokens.append(lemmatize_token(word, pos))

    return tuple(lemmatized_tokens)

    # n = 0
    # for word, lemma in zip(wordlst, wordlst_lem):
    #     if word != lemma:
    #         n += 1
    # print(
    #     f'Entries for which lemmatized form of word differs from the actual word: {n} words, {n/len(wordlst)*100:.2f}%')
    # print('-'*79)


# @cache_to_mem
def get_is_content(taglst: tuple, content_pos=(wordnet.ADJ, wordnet.VERB, wordnet.NOUN, wordnet.ADV)):
    """
    Given list of POS tags, return list of 1 - content word, 0 - not content word
    """
    return tuple(int(get_wordnet_pos(tag) in content_pos) for tag in taglst)


def get_wordnet_pos(treebank_tag):
    if treebank_tag.startswith('J'):
        return wordnet.ADJ
    elif treebank_tag.startswith('V'):
        return wordnet.VERB
    elif treebank_tag.startswith('N'):
        return wordnet.NOUN
    elif treebank_tag.startswith('R'):
        return wordnet.ADV
    else:
        return ''
